# Copyright 2011 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections.abc
import copy
import math
from typing import Any
from unittest import mock

import ddt
from oslotest import base as test_base
import testscenarios

from oslo_utils import strutils
from oslo_utils import units

load_tests = testscenarios.load_tests_apply_scenarios


class StrUtilsTest(test_base.BaseTestCase):
    def test_bool_bool_from_string(self):
        self.assertTrue(strutils.bool_from_string(True))
        self.assertFalse(strutils.bool_from_string(False))

    def test_bool_bool_from_string_default(self):
        self.assertTrue(strutils.bool_from_string('', default=True))
        self.assertFalse(strutils.bool_from_string('wibble', default=False))

    def _test_bool_from_string(self, c):
        self.assertTrue(strutils.bool_from_string(c('true')))
        self.assertTrue(strutils.bool_from_string(c('TRUE')))
        self.assertTrue(strutils.bool_from_string(c('on')))
        self.assertTrue(strutils.bool_from_string(c('On')))
        self.assertTrue(strutils.bool_from_string(c('yes')))
        self.assertTrue(strutils.bool_from_string(c('YES')))
        self.assertTrue(strutils.bool_from_string(c('yEs')))
        self.assertTrue(strutils.bool_from_string(c('1')))
        self.assertTrue(strutils.bool_from_string(c('T')))
        self.assertTrue(strutils.bool_from_string(c('t')))
        self.assertTrue(strutils.bool_from_string(c('Y')))
        self.assertTrue(strutils.bool_from_string(c('y')))

        self.assertFalse(strutils.bool_from_string(c('false')))
        self.assertFalse(strutils.bool_from_string(c('FALSE')))
        self.assertFalse(strutils.bool_from_string(c('off')))
        self.assertFalse(strutils.bool_from_string(c('OFF')))
        self.assertFalse(strutils.bool_from_string(c('no')))
        self.assertFalse(strutils.bool_from_string(c('0')))
        self.assertFalse(strutils.bool_from_string(c('42')))
        self.assertFalse(
            strutils.bool_from_string(c('This should not be True'))
        )
        self.assertFalse(strutils.bool_from_string(c('F')))
        self.assertFalse(strutils.bool_from_string(c('f')))
        self.assertFalse(strutils.bool_from_string(c('N')))
        self.assertFalse(strutils.bool_from_string(c('n')))

        # Whitespace should be stripped
        self.assertTrue(strutils.bool_from_string(c(' 1 ')))
        self.assertTrue(strutils.bool_from_string(c(' true ')))
        self.assertFalse(strutils.bool_from_string(c(' 0 ')))
        self.assertFalse(strutils.bool_from_string(c(' false ')))

    def test_bool_from_string(self):
        self._test_bool_from_string(lambda s: s)

    def test_unicode_bool_from_string(self):
        self._test_bool_from_string(str)
        self.assertFalse(strutils.bool_from_string('使用', strict=False))

        exc = self.assertRaises(
            ValueError, strutils.bool_from_string, '使用', strict=True
        )
        expected_msg = (
            "Unrecognized value '使用', acceptable values are:"
            " '0', '1', 'f', 'false', 'n', 'no', 'off', 'on',"
            " 't', 'true', 'y', 'yes'"
        )
        self.assertEqual(expected_msg, str(exc))

    def test_other_bool_from_string(self):
        self.assertFalse(strutils.bool_from_string(None))
        self.assertFalse(strutils.bool_from_string(mock.Mock()))

    def test_int_bool_from_string(self):
        self.assertTrue(strutils.bool_from_string(1))

        self.assertFalse(strutils.bool_from_string(-1))
        self.assertFalse(strutils.bool_from_string(0))
        self.assertFalse(strutils.bool_from_string(2))

    def test_strict_bool_from_string(self):
        # None isn't allowed in strict mode
        exc = self.assertRaises(
            ValueError, strutils.bool_from_string, None, strict=True
        )
        expected_msg = (
            "Unrecognized value 'None', acceptable values are:"
            " '0', '1', 'f', 'false', 'n', 'no', 'off', 'on',"
            " 't', 'true', 'y', 'yes'"
        )
        self.assertEqual(expected_msg, str(exc))

        # Unrecognized strings aren't allowed
        self.assertFalse(strutils.bool_from_string('Other', strict=False))
        exc = self.assertRaises(
            ValueError, strutils.bool_from_string, 'Other', strict=True
        )
        expected_msg = (
            "Unrecognized value 'Other', acceptable values are:"
            " '0', '1', 'f', 'false', 'n', 'no', 'off', 'on',"
            " 't', 'true', 'y', 'yes'"
        )
        self.assertEqual(expected_msg, str(exc))

        # Unrecognized numbers aren't allowed
        exc = self.assertRaises(
            ValueError, strutils.bool_from_string, 2, strict=True
        )
        expected_msg = (
            "Unrecognized value '2', acceptable values are:"
            " '0', '1', 'f', 'false', 'n', 'no', 'off', 'on',"
            " 't', 'true', 'y', 'yes'"
        )
        self.assertEqual(expected_msg, str(exc))

        # False-like values are allowed
        self.assertFalse(strutils.bool_from_string('f', strict=True))
        self.assertFalse(strutils.bool_from_string('false', strict=True))
        self.assertFalse(strutils.bool_from_string('off', strict=True))
        self.assertFalse(strutils.bool_from_string('n', strict=True))
        self.assertFalse(strutils.bool_from_string('no', strict=True))
        self.assertFalse(strutils.bool_from_string('0', strict=True))

        self.assertTrue(strutils.bool_from_string('1', strict=True))

        # Avoid font-similarity issues (one looks like lowercase-el, zero like
        # oh, etc...)
        for char in ('O', 'o', 'L', 'l', 'I', 'i'):
            self.assertRaises(
                ValueError, strutils.bool_from_string, char, strict=True
            )

    def test_int_from_bool_as_string(self):
        self.assertEqual(1, strutils.int_from_bool_as_string(True))
        self.assertEqual(0, strutils.int_from_bool_as_string(False))

    def test_is_valid_boolstr(self):
        self.assertTrue(strutils.is_valid_boolstr('true'))
        self.assertTrue(strutils.is_valid_boolstr('false'))
        self.assertTrue(strutils.is_valid_boolstr('yes'))
        self.assertTrue(strutils.is_valid_boolstr('no'))
        self.assertTrue(strutils.is_valid_boolstr('y'))
        self.assertTrue(strutils.is_valid_boolstr('n'))
        self.assertTrue(strutils.is_valid_boolstr('1'))
        self.assertTrue(strutils.is_valid_boolstr('0'))
        self.assertTrue(strutils.is_valid_boolstr(1))
        self.assertTrue(strutils.is_valid_boolstr(0))

        self.assertFalse(strutils.is_valid_boolstr('maybe'))
        self.assertFalse(strutils.is_valid_boolstr('only on tuesdays'))

    def test_slugify(self):
        to_slug = strutils.to_slug
        self.assertRaises(TypeError, to_slug, True)
        self.assertEqual("hello", to_slug("hello"))
        self.assertEqual("two-words", to_slug("Two Words"))
        self.assertEqual("ma-any-spa-ce-es", to_slug("Ma-any\t spa--ce- es"))
        self.assertEqual("excamation", to_slug("exc!amation!"))
        self.assertEqual("ampserand", to_slug("&ampser$and"))
        self.assertEqual("ju5tnum8er", to_slug("ju5tnum8er"))
        self.assertEqual("strip-", to_slug(" strip - "))
        self.assertEqual("perche", to_slug(b"perch\xc3\xa9"))
        self.assertEqual("strange", to_slug("\x80strange", errors="ignore"))


class StringToBytesTest(test_base.BaseTestCase):
    _unit_system = [
        ('si', dict(unit_system='SI')),
        ('iec', dict(unit_system='IEC')),
        ('mixed', dict(unit_system='mixed')),
        ('invalid_unit_system', dict(unit_system='KKK', assert_error=True)),
    ]

    _sign = [
        ('no_sign', dict(sign='')),
        ('positive', dict(sign='+')),
        ('negative', dict(sign='-')),
        ('invalid_sign', dict(sign='~', assert_error=True)),
    ]

    _magnitude = [
        ('integer', dict(magnitude='79')),
        ('decimal', dict(magnitude='7.9')),
        ('decimal_point_start', dict(magnitude='.9')),
        ('decimal_point_end', dict(magnitude='79.', assert_error=True)),
        ('invalid_literal', dict(magnitude='7.9.9', assert_error=True)),
        ('garbage_value', dict(magnitude='asdf', assert_error=True)),
    ]

    _unit_prefix = [
        ('no_unit_prefix', dict(unit_prefix='')),
        ('k', dict(unit_prefix='k')),
        ('K', dict(unit_prefix='K')),
        ('M', dict(unit_prefix='M')),
        ('G', dict(unit_prefix='G')),
        ('T', dict(unit_prefix='T')),
        ('P', dict(unit_prefix='P')),
        ('E', dict(unit_prefix='E')),
        ('Z', dict(unit_prefix='Z')),
        ('Y', dict(unit_prefix='Y')),
        ('R', dict(unit_prefix='R')),
        ('Q', dict(unit_prefix='Q')),
        ('Ki', dict(unit_prefix='Ki')),
        ('Mi', dict(unit_prefix='Mi')),
        ('Gi', dict(unit_prefix='Gi')),
        ('Ti', dict(unit_prefix='Ti')),
        ('Pi', dict(unit_prefix='Pi')),
        ('Ei', dict(unit_prefix='Ei')),
        ('Zi', dict(unit_prefix='Zi')),
        ('Yi', dict(unit_prefix='Yi')),
        ('Ri', dict(unit_prefix='Ri')),
        ('Qi', dict(unit_prefix='Qi')),
        ('invalid_unit_prefix', dict(unit_prefix='B', assert_error=True)),
    ]

    _unit_suffix = [
        ('b', dict(unit_suffix='b')),
        ('bit', dict(unit_suffix='bit')),
        ('B', dict(unit_suffix='B')),
        ('invalid_unit_suffix', dict(unit_suffix='Kg', assert_error=True)),
    ]

    _return_int = [
        ('return_dec', dict(return_int=False)),
        ('return_int', dict(return_int=True)),
    ]

    @classmethod
    def generate_scenarios(cls):
        cls.scenarios = testscenarios.multiply_scenarios(
            cls._unit_system,
            cls._sign,
            cls._magnitude,
            cls._unit_prefix,
            cls._unit_suffix,
            cls._return_int,
        )

    def test_string_to_bytes(self):
        def _get_quantity(sign, magnitude, unit_suffix):
            res = float(f'{sign}{magnitude}')
            if unit_suffix in ['b', 'bit']:
                res /= 8
            return res

        def _get_constant(unit_prefix, unit_system):
            if not unit_prefix:
                return 1
            elif unit_system == 'SI':
                res = getattr(units, unit_prefix)
            elif unit_system == 'IEC':
                if unit_prefix.endswith('i'):
                    res = getattr(units, unit_prefix)
                else:
                    res = getattr(units, f'{unit_prefix}i')
            elif unit_system == 'mixed':
                # Note: this will return 'i' units as power-of-two,
                # and other units as power-of-ten.  Additionally, for
                # compatability a "K" is interpreted as "k" in mixed
                # mode
                if unit_prefix == 'K':
                    unit_prefix = 'k'
                res = getattr(units, unit_prefix)
            return res

        text = ''.join(
            [self.sign, self.magnitude, self.unit_prefix, self.unit_suffix]
        )
        err_si = self.unit_system == 'SI' and (
            self.unit_prefix == 'K' or self.unit_prefix.endswith('i')
        )
        err_iec = self.unit_system == 'IEC' and self.unit_prefix == 'k'
        if getattr(self, 'assert_error', False) or err_si or err_iec:
            self.assertRaises(
                ValueError,
                strutils.string_to_bytes,
                text,
                unit_system=self.unit_system,
                return_int=self.return_int,
            )
            return
        quantity = _get_quantity(self.sign, self.magnitude, self.unit_suffix)
        constant = _get_constant(self.unit_prefix, self.unit_system)
        expected = quantity * constant
        actual = strutils.string_to_bytes(
            text, unit_system=self.unit_system, return_int=self.return_int
        )
        if self.return_int:
            self.assertEqual(actual, int(math.ceil(expected)))
        else:
            self.assertAlmostEqual(actual, expected)


StringToBytesTest.generate_scenarios()


class MaskPasswordTestCase(test_base.BaseTestCase):
    def test_namespace_objects(self):
        payload = """
        Namespace(passcode='', username='', password='my"password',
        profile='', verify=None, token='')
        """
        expected = """
        Namespace(passcode='', username='', password='***',
        profile='', verify=None, token='***')
        """
        self.assertEqual(expected, strutils.mask_password(payload))

    def test_sanitize_keys(self):
        lowered = [k.lower() for k in strutils._SANITIZE_KEYS]
        message = "The _SANITIZE_KEYS must all be lowercase."
        self.assertEqual(strutils._SANITIZE_KEYS, lowered, message)

    def test_json(self):
        # Test 'adminPass' w/o spaces
        payload = """{'adminPass':'TL0EfN33'}"""
        expected = """{'adminPass':'***'}"""
        self.assertEqual(expected, strutils.mask_password(payload))
        # Test 'adminPass' with spaces
        payload = """{ 'adminPass' : 'TL0EfN33' }"""
        expected = """{ 'adminPass' : '***' }"""
        self.assertEqual(expected, strutils.mask_password(payload))
        # Test 'admin_pass' w/o spaces
        payload = """{'admin_pass':'TL0EfN33'}"""
        expected = """{'admin_pass':'***'}"""
        self.assertEqual(expected, strutils.mask_password(payload))
        # Test 'admin_pass' with spaces
        payload = """{ 'admin_pass' : 'TL0EfN33' }"""
        expected = """{ 'admin_pass' : '***' }"""
        self.assertEqual(expected, strutils.mask_password(payload))
        # Test 'admin_password' w/o spaces
        payload = """{'admin_password':'TL0EfN33'}"""
        expected = """{'admin_password':'***'}"""
        self.assertEqual(expected, strutils.mask_password(payload))
        # Test 'admin_password' with spaces
        payload = """{ 'admin_password' : 'TL0EfN33' }"""
        expected = """{ 'admin_password' : '***' }"""
        self.assertEqual(expected, strutils.mask_password(payload))
        # Test 'password' w/o spaces
        payload = """{'password':'TL0EfN33'}"""
        expected = """{'password':'***'}"""
        self.assertEqual(expected, strutils.mask_password(payload))
        # Test 'password' with spaces
        payload = """{ 'password' : 'TL0EfN33' }"""
        expected = """{ 'password' : '***' }"""
        self.assertEqual(expected, strutils.mask_password(payload))
        # Test 'auth_password' w/o spaces
        payload = """{'auth_password':'TL0EfN33'}"""
        expected = """{'auth_password':'***'}"""
        self.assertEqual(expected, strutils.mask_password(payload))
        # Test 'auth_password' with spaces
        payload = """{ 'auth_password' : 'TL0EfN33' }"""
        expected = """{ 'auth_password' : '***' }"""
        self.assertEqual(expected, strutils.mask_password(payload))
        # Test 'secret_uuid' w/o spaces
        payload = """{'secret_uuid':'myuuid'}"""
        expected = """{'secret_uuid':'***'}"""
        self.assertEqual(expected, strutils.mask_password(payload))
        # Test 'secret_uuid' with spaces
        payload = """{ 'secret_uuid' : 'myuuid' }"""
        expected = """{ 'secret_uuid' : '***' }"""
        self.assertEqual(expected, strutils.mask_password(payload))
        # Test 'token' w/o spaces
        payload = """{'token':'token'}"""
        expected = """{'token':'***'}"""
        self.assertEqual(expected, strutils.mask_password(payload))
        # Test 'token' with spaces
        payload = """{ 'token' : 'token' }"""
        expected = """{ 'token' : '***' }"""
        self.assertEqual(expected, strutils.mask_password(payload))
        # Test 'fernetkey'
        payload = """{ 'fernetkey' : 'token' }"""
        expected = """{ 'fernetkey' : '***' }"""
        self.assertEqual(expected, strutils.mask_password(payload))
        # Test 'FernetKey'
        payload = """{ 'FernetKey' : 'token' }"""
        expected = """{ 'FernetKey' : '***' }"""
        self.assertEqual(expected, strutils.mask_password(payload))
        # Test 'sslkey'
        payload = """{ 'sslkey' : 'token' }"""
        expected = """{ 'sslkey' : '***' }"""
        self.assertEqual(expected, strutils.mask_password(payload))
        # Test 'SslKey'
        payload = """{ 'SslKey' : 'token' }"""
        expected = """{ 'SslKey' : '***' }"""
        self.assertEqual(expected, strutils.mask_password(payload))
        # Test 'passphrase'
        payload = """{ 'passphrase' : 'token' }"""
        expected = """{ 'passphrase' : '***' }"""
        self.assertEqual(expected, strutils.mask_password(payload))
        # Test 'PassPhrase'
        payload = """{ 'PassPhrase' : 'token' }"""
        expected = """{ 'PassPhrase' : '***' }"""
        self.assertEqual(expected, strutils.mask_password(payload))
        # Some real-life cases
        # Test 'KeystoneFernetKey1'
        payload = """{ 'KeystoneFernetKey1' : 'token' }"""
        expected = """{ 'KeystoneFernetKey1' : '***' }"""
        self.assertEqual(expected, strutils.mask_password(payload))
        # Test 'OctaviaCaKeyPassword'
        payload = """{ 'OctaviaCaKeyPassword' : 'token' }"""
        expected = """{ 'OctaviaCaKeyPassword' : '***' }"""
        self.assertEqual(expected, strutils.mask_password(payload))
        # Test 'OctaviaCaKeyPassphrase'
        payload = """{ 'OctaviaCaKeyPassphrase' : 'token' }"""
        expected = """{ 'OctaviaCaKeyPassphrase' : '***' }"""
        self.assertEqual(expected, strutils.mask_password(payload))

    def test_xml(self):
        # Test 'adminPass' w/o spaces
        payload = """<adminPass>TL0EfN33</adminPass>"""
        expected = """<adminPass>***</adminPass>"""
        self.assertEqual(expected, strutils.mask_password(payload))
        # Test 'adminPass' with spaces
        payload = """<adminPass>
                        TL0EfN33
                     </adminPass>"""
        expected = """<adminPass>***</adminPass>"""
        self.assertEqual(expected, strutils.mask_password(payload))
        # Test 'admin_pass' w/o spaces
        payload = """<admin_pass>TL0EfN33</admin_pass>"""
        expected = """<admin_pass>***</admin_pass>"""
        self.assertEqual(expected, strutils.mask_password(payload))
        # Test 'admin_pass' with spaces
        payload = """<admin_pass>
                        TL0EfN33
                     </admin_pass>"""
        expected = """<admin_pass>***</admin_pass>"""
        self.assertEqual(expected, strutils.mask_password(payload))
        # Test 'admin_password' w/o spaces
        payload = """<admin_password>TL0EfN33</admin_password>"""
        expected = """<admin_password>***</admin_password>"""
        self.assertEqual(expected, strutils.mask_password(payload))
        # Test 'admin_password' with spaces
        payload = """<admin_password>
                        TL0EfN33
                     </admin_password>"""
        expected = """<admin_password>***</admin_password>"""
        self.assertEqual(expected, strutils.mask_password(payload))
        # Test 'password' w/o spaces
        payload = """<password>TL0EfN33</password>"""
        expected = """<password>***</password>"""
        self.assertEqual(expected, strutils.mask_password(payload))
        # Test 'password' with spaces
        payload = """<password>
                        TL0EfN33
                     </password>"""
        expected = """<password>***</password>"""
        self.assertEqual(expected, strutils.mask_password(payload))
        # Test 'Password1' - case-insensitive + number
        payload = """<Password1>TL0EfN33</Password1>"""
        expected = """<Password1>***</Password1>"""
        self.assertEqual(expected, strutils.mask_password(payload))

    def test_xml_attribute(self):
        # Test 'adminPass' w/o spaces
        payload = """adminPass='TL0EfN33'"""
        expected = """adminPass='***'"""
        self.assertEqual(expected, strutils.mask_password(payload))
        # Test 'adminPass' with spaces
        payload = """adminPass = 'TL0EfN33'"""
        expected = """adminPass = '***'"""
        self.assertEqual(expected, strutils.mask_password(payload))
        # Test 'adminPass' with double quotes
        payload = """adminPass = "TL0EfN33\""""
        expected = """adminPass = "***\""""
        self.assertEqual(expected, strutils.mask_password(payload))
        # Test 'admin_pass' w/o spaces
        payload = """admin_pass='TL0EfN33'"""
        expected = """admin_pass='***'"""
        self.assertEqual(expected, strutils.mask_password(payload))
        # Test 'admin_pass' with spaces
        payload = """admin_pass = 'TL0EfN33'"""
        expected = """admin_pass = '***'"""
        self.assertEqual(expected, strutils.mask_password(payload))
        # Test 'admin_pass' with double quotes
        payload = """admin_pass = "TL0EfN33\""""
        expected = """admin_pass = "***\""""
        self.assertEqual(expected, strutils.mask_password(payload))
        # Test 'admin_password' w/o spaces
        payload = """admin_password='TL0EfN33'"""
        expected = """admin_password='***'"""
        self.assertEqual(expected, strutils.mask_password(payload))
        # Test 'admin_password' with spaces
        payload = """admin_password = 'TL0EfN33'"""
        expected = """admin_password = '***'"""
        self.assertEqual(expected, strutils.mask_password(payload))
        # Test 'admin_password' with double quotes
        payload = """admin_password = "TL0EfN33\""""
        expected = """admin_password = "***\""""
        self.assertEqual(expected, strutils.mask_password(payload))
        # Test 'password' w/o spaces
        payload = """password='TL0EfN33'"""
        expected = """password='***'"""
        self.assertEqual(expected, strutils.mask_password(payload))
        # Test 'password' with spaces
        payload = """password = 'TL0EfN33'"""
        expected = """password = '***'"""
        self.assertEqual(expected, strutils.mask_password(payload))
        # Test 'password' with double quotes
        payload = """password = "TL0EfN33\""""
        expected = """password = "***\""""
        self.assertEqual(expected, strutils.mask_password(payload))

    def test_json_message(self):
        payload = """body: {"changePassword": {"adminPass": "1234567"}}"""
        expected = """body: {"changePassword": {"adminPass": "***"}}"""
        self.assertEqual(expected, strutils.mask_password(payload))
        payload = """body: {"rescue": {"admin_pass": "1234567"}}"""
        expected = """body: {"rescue": {"admin_pass": "***"}}"""
        self.assertEqual(expected, strutils.mask_password(payload))
        payload = """body: {"rescue": {"admin_password": "1234567"}}"""
        expected = """body: {"rescue": {"admin_password": "***"}}"""
        self.assertEqual(expected, strutils.mask_password(payload))
        payload = """body: {"rescue": {"password": "1234567"}}"""
        expected = """body: {"rescue": {"password": "***"}}"""
        self.assertEqual(expected, strutils.mask_password(payload))

    def test_xml_message(self):
        payload = """<?xml version="1.0" encoding="UTF-8"?>
<rebuild
    xmlns="http://docs.openstack.org/compute/api/v1.1"
    name="foobar"
    imageRef="http://openstack.example.com/v1.1/32278/images/70a599e0-31e7"
    accessIPv4="1.2.3.4"
    accessIPv6="fe80::100"
    adminPass="seekr3t">
  <metadata>
    <meta key="My Server Name">Apache1</meta>
  </metadata>
</rebuild>"""
        expected = """<?xml version="1.0" encoding="UTF-8"?>
<rebuild
    xmlns="http://docs.openstack.org/compute/api/v1.1"
    name="foobar"
    imageRef="http://openstack.example.com/v1.1/32278/images/70a599e0-31e7"
    accessIPv4="1.2.3.4"
    accessIPv6="fe80::100"
    adminPass="***">
  <metadata>
    <meta key="My Server Name">Apache1</meta>
  </metadata>
</rebuild>"""
        self.assertEqual(expected, strutils.mask_password(payload))
        payload = """<?xml version="1.0" encoding="UTF-8"?>
<rescue xmlns="http://docs.openstack.org/compute/api/v1.1"
    admin_pass="MySecretPass"/>"""
        expected = """<?xml version="1.0" encoding="UTF-8"?>
<rescue xmlns="http://docs.openstack.org/compute/api/v1.1"
    admin_pass="***"/>"""
        self.assertEqual(expected, strutils.mask_password(payload))
        payload = """<?xml version="1.0" encoding="UTF-8"?>
<rescue xmlns="http://docs.openstack.org/compute/api/v1.1"
    admin_password="MySecretPass"/>"""
        expected = """<?xml version="1.0" encoding="UTF-8"?>
<rescue xmlns="http://docs.openstack.org/compute/api/v1.1"
    admin_password="***"/>"""
        self.assertEqual(expected, strutils.mask_password(payload))
        payload = """<?xml version="1.0" encoding="UTF-8"?>
<rescue xmlns="http://docs.openstack.org/compute/api/v1.1"
    password="MySecretPass"/>"""
        expected = """<?xml version="1.0" encoding="UTF-8"?>
<rescue xmlns="http://docs.openstack.org/compute/api/v1.1"
    password="***"/>"""
        self.assertEqual(expected, strutils.mask_password(payload))

    def test_mask_password(self):
        payload = "test = 'password'  :   'aaaaaa'"
        expected = "test = 'password'  :   '111'"
        self.assertEqual(
            expected, strutils.mask_password(payload, secret='111')
        )

        payload = 'mysqld --password "aaaaaa"'
        expected = 'mysqld --password "****"'
        self.assertEqual(
            expected, strutils.mask_password(payload, secret='****')
        )

        payload = 'mysqld --password aaaaaa'
        expected = 'mysqld --password ???'
        self.assertEqual(
            expected, strutils.mask_password(payload, secret='???')
        )

        payload = 'mysqld --password = "aaaaaa"'
        expected = 'mysqld --password = "****"'
        self.assertEqual(
            expected, strutils.mask_password(payload, secret='****')
        )

        payload = "mysqld --password = 'aaaaaa'"
        expected = "mysqld --password = '****'"
        self.assertEqual(
            expected, strutils.mask_password(payload, secret='****')
        )

        payload = "mysqld --password = aaaaaa"
        expected = "mysqld --password = ****"
        self.assertEqual(
            expected, strutils.mask_password(payload, secret='****')
        )

        payload = "test = password =   aaaaaa"
        expected = "test = password =   111"
        self.assertEqual(
            expected, strutils.mask_password(payload, secret='111')
        )

        payload = "test = password=   aaaaaa"
        expected = "test = password=   111"
        self.assertEqual(
            expected, strutils.mask_password(payload, secret='111')
        )

        payload = "test = password =aaaaaa"
        expected = "test = password =111"
        self.assertEqual(
            expected, strutils.mask_password(payload, secret='111')
        )

        payload = "test = password=aaaaaa"
        expected = "test = password=111"
        self.assertEqual(
            expected, strutils.mask_password(payload, secret='111')
        )

        payload = 'test = "original_password" : "aaaaaaaaa"'
        expected = 'test = "original_password" : "***"'
        self.assertEqual(expected, strutils.mask_password(payload))

        payload = 'test = "param1" : "value"'
        expected = 'test = "param1" : "value"'
        self.assertEqual(expected, strutils.mask_password(payload))

        payload = 'test = "original_password" : "aaaaa"aaaa"'
        expected = 'test = "original_password" : "***"'
        self.assertEqual(expected, strutils.mask_password(payload))

        payload = """{'adminPass':'TL0EfN33'}"""
        payload = str(payload)
        expected = """{'adminPass':'***'}"""
        self.assertEqual(expected, strutils.mask_password(payload))

        payload = """{'adminPass':'TL0E'fN33'}"""
        payload = str(payload)
        expected = """{'adminPass':'***'}"""
        self.assertEqual(expected, strutils.mask_password(payload))

        payload = """{'token':'mytoken'}"""
        payload = str(payload)
        expected = """{'token':'***'}"""
        self.assertEqual(expected, strutils.mask_password(payload))

        payload = (
            "test = 'node.session.auth.password','-v','TL0EfN33','nomask'"
        )
        expected = "test = 'node.session.auth.password','-v','***','nomask'"
        self.assertEqual(expected, strutils.mask_password(payload))

        payload = (
            "test = 'node.session.auth.password', '--password', "
            "'TL0EfN33', 'nomask'"
        )
        expected = (
            "test = 'node.session.auth.password', '--password', "
            "'***', 'nomask'"
        )
        self.assertEqual(expected, strutils.mask_password(payload))

        payload = (
            "test = 'node.session.auth.password', '--password', 'TL0EfN33'"
        )
        expected = "test = 'node.session.auth.password', '--password', '***'"
        self.assertEqual(expected, strutils.mask_password(payload))

        payload = "test = node.session.auth.password -v TL0EfN33 nomask"
        expected = "test = node.session.auth.password -v *** nomask"
        self.assertEqual(expected, strutils.mask_password(payload))

        payload = (
            "test = node.session.auth.password --password TL0EfN33 nomask"
        )
        expected = "test = node.session.auth.password --password *** nomask"
        self.assertEqual(expected, strutils.mask_password(payload))

        payload = "test = node.session.auth.password --password TL0EfN33"
        expected = "test = node.session.auth.password --password ***"
        self.assertEqual(expected, strutils.mask_password(payload))

        payload = "test = cmd --password my\xe9\x80\x80pass"
        expected = "test = cmd --password ***"
        self.assertEqual(expected, strutils.mask_password(payload))


class TestMapping(collections.abc.Mapping[str, Any]):
    """Test class for non-dict mappings"""

    def __init__(self):
        super().__init__()
        self.data: dict[str, Any] = {
            'password': 'shhh',
            'foo': 'bar',
        }

    def __getitem__(self, key):
        return self.data[key]

    def __iter__(self):
        return self.data.__iter__()

    def __len__(self):
        return len(self.data)


class NestedMapping(TestMapping):
    """Test class that contains an instance of TestMapping"""

    def __init__(self):
        super().__init__()
        self.data = {'nested': TestMapping()}


class MaskDictionaryPasswordTestCase(test_base.BaseTestCase):
    def test_dictionary(self):
        payload: dict[str, Any] = {'password': 'TL0EfN33'}
        expected: dict[str, Any] = {'password': '***'}
        self.assertEqual(expected, strutils.mask_dict_password(payload))

        payload = {'password': 'TL0Ef"N33'}
        expected = {'password': '***'}
        self.assertEqual(expected, strutils.mask_dict_password(payload))

        payload = {'user': 'admin', 'password': 'TL0EfN33'}
        expected = {'user': 'admin', 'password': '***'}
        self.assertEqual(expected, strutils.mask_dict_password(payload))

        payload = {
            'strval': 'somestring',
            'dictval': {'user': 'admin', 'password': 'TL0EfN33'},
        }
        expected = {
            'strval': 'somestring',
            'dictval': {'user': 'admin', 'password': '***'},
        }
        self.assertEqual(expected, strutils.mask_dict_password(payload))

        payload = {
            'strval': '--password abc',
            'dont_change': 'this is fine',
            'dictval': {'user': 'admin', 'password': b'TL0EfN33'},
        }
        expected = {
            'strval': '--password ***',
            'dont_change': 'this is fine',
            'dictval': {'user': 'admin', 'password': '***'},
        }
        self.assertEqual(expected, strutils.mask_dict_password(payload))

        payload = {'ipmi_password': 'KeDrahishvowphyecMornEm0or('}
        expected = {'ipmi_password': '***'}
        self.assertEqual(expected, strutils.mask_dict_password(payload))

        payload = {'passwords': {'KeystoneFernetKey1': 'c5FijjS'}}
        expected = {'passwords': {'KeystoneFernetKey1': '***'}}
        self.assertEqual(expected, strutils.mask_dict_password(payload))

        payload = {'passwords': {'keystonecredential0': 'c5FijjS'}}
        expected = {'passwords': {'keystonecredential0': '***'}}
        self.assertEqual(expected, strutils.mask_dict_password(payload))

    def test_do_no_harm(self):
        self.assertEqual({}, strutils.mask_dict_password({}))

        payload = {'somekey': 'somevalue', 'anotherkey': 'anothervalue'}
        expected = {'somekey': 'somevalue', 'anotherkey': 'anothervalue'}
        self.assertEqual(expected, strutils.mask_dict_password(payload))

    def test_do_an_int(self):
        payload = {}
        payload[1] = 2
        expected = payload.copy()
        self.assertEqual(expected, strutils.mask_dict_password(payload))

    def test_mask_values(self):
        payload = {'somekey': 'test = cmd --password my\xe9\x80\x80pass'}
        expected = {'somekey': 'test = cmd --password ***'}
        self.assertEqual(expected, strutils.mask_dict_password(payload))

    def test_other_non_str_values(self):
        payload = {
            'password': 'DK0PK1AK3',
            'bool': True,
            'dict': {'cat': 'meow', 'password': "*aa38skdjf"},
            'float': 0.1,
            'int': 123,
            'list': [1, 2],
            'none': None,
            'str': 'foo',
        }
        expected = {
            'password': '***',
            'bool': True,
            'dict': {'cat': 'meow', 'password': '***'},
            'float': 0.1,
            'int': 123,
            'list': [1, 2],
            'none': None,
            'str': 'foo',
        }
        self.assertEqual(expected, strutils.mask_dict_password(payload))

    def test_argument_untouched(self):
        """Make sure that the argument passed in is not modified"""
        payload = {
            'password': 'DK0PK1AK3',
            'bool': True,
            'dict': {'cat': 'meow', 'password': "*aa38skdjf"},
            'float': 0.1,
            'int': 123,
            'list': [1, 2],
            'none': None,
            'str': 'foo',
        }
        pristine = copy.deepcopy(payload)
        # Send the payload into the function, to see if it gets modified
        strutils.mask_dict_password(payload)
        self.assertEqual(pristine, payload)

    def test_non_dict(self):
        expected = {
            'password': '***',
            'foo': 'bar',
        }
        payload = TestMapping()
        self.assertEqual(expected, strutils.mask_dict_password(payload))

    def test_nested_non_dict(self):
        expected = {
            'nested': {
                'password': '***',
                'foo': 'bar',
            }
        }
        payload = NestedMapping()
        self.assertEqual(expected, strutils.mask_dict_password(payload))


class IsIntLikeTestCase(test_base.BaseTestCase):
    def test_is_int_like_true(self):
        self.assertTrue(strutils.is_int_like(1))
        self.assertTrue(strutils.is_int_like("1"))
        self.assertTrue(strutils.is_int_like("514"))
        self.assertTrue(strutils.is_int_like("0"))

    def test_is_int_like_false(self):
        self.assertFalse(strutils.is_int_like(1.1))
        self.assertFalse(strutils.is_int_like("1.1"))
        self.assertFalse(strutils.is_int_like("1.1.1"))
        self.assertFalse(strutils.is_int_like(None))
        self.assertFalse(strutils.is_int_like("0."))
        self.assertFalse(strutils.is_int_like("aaaaaa"))
        self.assertFalse(strutils.is_int_like("...."))
        self.assertFalse(strutils.is_int_like("1g"))
        self.assertFalse(
            strutils.is_int_like("0cc3346e-9fef-4445-abe6-5d2b2690ec64")
        )
        self.assertFalse(strutils.is_int_like("a1"))
        # NOTE(viktors): 12e3 - is a float number
        self.assertFalse(strutils.is_int_like("12e3"))
        # NOTE(viktors): Check integer numbers with base not 10
        self.assertFalse(strutils.is_int_like("0o51"))
        self.assertFalse(strutils.is_int_like("0xDEADBEEF"))


class StringLengthTestCase(test_base.BaseTestCase):
    def test_check_string_length(self):
        self.assertIsNone(
            strutils.check_string_length('test', 'name', max_length=255)  # type: ignore
        )
        self.assertRaises(
            ValueError, strutils.check_string_length, '', 'name', min_length=1
        )
        self.assertRaises(
            ValueError,
            strutils.check_string_length,
            'a' * 256,
            'name',
            max_length=255,
        )
        self.assertRaises(
            TypeError, strutils.check_string_length, 11, 'name', max_length=255
        )
        self.assertRaises(
            TypeError,
            strutils.check_string_length,
            dict(),
            'name',
            max_length=255,
        )

    def test_check_string_length_noname(self):
        self.assertIsNone(strutils.check_string_length('test', max_length=255))  # type: ignore
        self.assertRaises(
            ValueError, strutils.check_string_length, '', min_length=1
        )
        self.assertRaises(
            ValueError, strutils.check_string_length, 'a' * 256, max_length=255
        )
        self.assertRaises(
            TypeError, strutils.check_string_length, 11, max_length=255
        )
        self.assertRaises(
            TypeError, strutils.check_string_length, dict(), max_length=255
        )


class SplitPathTestCase(test_base.BaseTestCase):
    def test_split_path_failed(self):
        self.assertRaises(ValueError, strutils.split_path, '')
        self.assertRaises(ValueError, strutils.split_path, '/')
        self.assertRaises(ValueError, strutils.split_path, '//')
        self.assertRaises(ValueError, strutils.split_path, '//a')
        self.assertRaises(ValueError, strutils.split_path, '/a/c')
        self.assertRaises(ValueError, strutils.split_path, '//c')
        self.assertRaises(ValueError, strutils.split_path, '/a/c/')
        self.assertRaises(ValueError, strutils.split_path, '/a//')
        self.assertRaises(ValueError, strutils.split_path, '/a', 2)
        self.assertRaises(ValueError, strutils.split_path, '/a', 2, 3)
        self.assertRaises(ValueError, strutils.split_path, '/a', 2, 3, True)
        self.assertRaises(ValueError, strutils.split_path, '/a/c/o/r', 3, 3)
        self.assertRaises(ValueError, strutils.split_path, '/a', 5, 4)

    def test_split_path_success(self):
        self.assertEqual(strutils.split_path('/a'), ['a'])
        self.assertEqual(strutils.split_path('/a/'), ['a'])
        self.assertEqual(strutils.split_path('/a/c', 2), ['a', 'c'])
        self.assertEqual(strutils.split_path('/a/c/o', 3), ['a', 'c', 'o'])
        self.assertEqual(
            strutils.split_path('/a/c/o/r', 3, 3, True), ['a', 'c', 'o/r']
        )
        self.assertEqual(
            strutils.split_path('/a/c', 2, 3, True), ['a', 'c', None]
        )
        self.assertEqual(strutils.split_path('/a/c/', 2), ['a', 'c'])
        self.assertEqual(strutils.split_path('/a/c/', 2, 3), ['a', 'c', ''])

    def test_split_path_invalid_path(self):
        try:
            strutils.split_path('o\nn e', 2)
        except ValueError as err:
            self.assertEqual(str(err), 'Invalid path: o%0An%20e')
        try:
            strutils.split_path('o\nn e', 2, 3, True)
        except ValueError as err:
            self.assertEqual(str(err), 'Invalid path: o%0An%20e')


class SplitByCommas(test_base.BaseTestCase):
    def test_not_closed_quotes(self):
        self.assertRaises(ValueError, strutils.split_by_commas, '"ab","b""')

    def test_no_comma_before_opening_quotes(self):
        self.assertRaises(ValueError, strutils.split_by_commas, '"ab""b"')

    def test_quote_inside_unquoted(self):
        self.assertRaises(ValueError, strutils.split_by_commas, 'a"b,cd')

    def check(self, expect, input):
        self.assertEqual(expect, strutils.split_by_commas(input))

    def test_plain(self):
        self.check(["a,b", "ac"], '"a,b",ac')

    def test_with_backslash_inside_quoted(self):
        self.check(
            ['abc"', 'de', 'fg,h', 'klm\\', '"nop'],
            r'"abc\"","de","fg,h","klm\\","\"nop"',
        )

    def test_with_backslash_inside_unquoted(self):
        self.check([r'a\bc', 'de'], r'a\bc,de')

    def test_with_escaped_quotes_in_row_inside_quoted(self):
        self.check(['a"b""c', 'd'], r'"a\"b\"\"c",d')


@ddt.ddt
class ValidateIntegerTestCase(test_base.BaseTestCase):
    @ddt.unpack
    @ddt.data(
        {"value": 42, "name": "answer", "output": 42},
        {"value": "42", "name": "answer", "output": 42},
        {
            "value": "7",
            "name": "lucky",
            "output": 7,
            "min_value": 7,
            "max_value": 8,
        },
        {
            "value": 7,
            "name": "lucky",
            "output": 7,
            "min_value": 6,
            "max_value": 7,
        },
        {"value": 300, "name": "Spartaaa!!!", "output": 300, "min_value": 300},
        {
            "value": "300",
            "name": "Spartaaa!!!",
            "output": 300,
            "max_value": 300,
        },
    )
    def test_valid_inputs(self, output, value, name, **kwargs):
        self.assertEqual(
            strutils.validate_integer(value, name, **kwargs), output
        )

    @ddt.unpack
    @ddt.data(
        {"value": "im-not-an-int", "name": ''},
        {"value": 3.14, "name": "Pie"},
        {
            "value": "299",
            "name": "Sparta no-show",
            "min_value": 300,
            "max_value": 300,
        },
        {"value": 55, "name": "doing 55 in a 54", "max_value": 54},
        {"value": chr(129), "name": "UnicodeError", "max_value": 1000},
    )
    def test_invalid_inputs(self, value, name, **kwargs):
        self.assertRaises(
            ValueError, strutils.validate_integer, value, name, **kwargs
        )



#    Copyright (C) 2012 Yahoo! Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import functools

from oslotest import base as test_base

from oslo_utils import reflection


RUNTIME_ERROR_CLASSES = [
    'RuntimeError',
    'Exception',
    'BaseException',
    'object',
]


def dummy_decorator(f):
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        return f(*args, **kwargs)

    return wrapper


def mere_function(a, b):
    pass


def function_with_defs(a, b, optional=None):
    pass


def function_with_kwargs(a, b, **kwargs):
    pass


class TestObject:
    def _hello(self):
        pass

    def hi(self):
        pass


class Class:
    def method(self, c, d):
        pass

    @staticmethod
    def static_method(e, f):
        pass

    @classmethod
    def class_method(cls, g, h):
        pass


class BadClass:
    def do_something(self):
        pass

    def __nonzero__(self):
        return False


class CallableClass:
    def __call__(self, i, j):
        pass


class ClassWithInit:
    def __init__(self, k, lll):
        pass


class MemberGetTest(test_base.BaseTestCase):
    def test_get_members_exclude_hidden(self):
        obj = TestObject()
        members = list(reflection.get_members(obj, exclude_hidden=True))
        self.assertEqual(1, len(members))

    def test_get_members_no_exclude_hidden(self):
        obj = TestObject()
        members = list(reflection.get_members(obj, exclude_hidden=False))
        self.assertGreater(len(members), 1)

    def test_get_members_names_exclude_hidden(self):
        obj = TestObject()
        members = list(reflection.get_member_names(obj, exclude_hidden=True))
        self.assertEqual(["hi"], members)

    def test_get_members_names_no_exclude_hidden(self):
        obj = TestObject()
        members = list(reflection.get_member_names(obj, exclude_hidden=False))
        members = [member for member in members if not member.startswith("__")]
        self.assertEqual(["_hello", "hi"], sorted(members))


class CallbackEqualityTest(test_base.BaseTestCase):
    def test_different_simple_callbacks(self):
        def a():
            pass

        def b():
            pass

        self.assertFalse(reflection.is_same_callback(a, b))

    def test_static_instance_callbacks(self):
        class A:
            @staticmethod
            def b(a, b, c):
                pass

        a = A()
        b = A()

        self.assertTrue(reflection.is_same_callback(a.b, b.b))

    def test_different_instance_callbacks(self):
        class A:
            def b(self):
                pass

            def __eq__(self, other):
                return True

            def __ne__(self, other):
                return not self.__eq__(other)

        b = A()
        c = A()

        self.assertFalse(reflection.is_same_callback(b.b, c.b))
        self.assertTrue(reflection.is_same_callback(b.b, b.b))


class BoundMethodTest(test_base.BaseTestCase):
    def test_baddy(self):
        b = BadClass()
        self.assertTrue(reflection.is_bound_method(b.do_something))

    def test_static_method(self):
        self.assertFalse(reflection.is_bound_method(Class.static_method))


class GetCallableNameTest(test_base.BaseTestCase):
    def test_mere_function(self):
        name = reflection.get_callable_name(mere_function)
        self.assertEqual('.'.join((__name__, 'mere_function')), name)

    def test_method(self):
        name = reflection.get_callable_name(Class.method)
        self.assertEqual('.'.join((__name__, 'Class', 'method')), name)

    def test_instance_method(self):
        name = reflection.get_callable_name(Class().method)
        self.assertEqual('.'.join((__name__, 'Class', 'method')), name)

    def test_static_method(self):
        name = reflection.get_callable_name(Class.static_method)
        self.assertEqual('.'.join((__name__, 'Class', 'static_method')), name)

    def test_class_method(self):
        name = reflection.get_callable_name(Class.class_method)
        self.assertEqual('.'.join((__name__, 'Class', 'class_method')), name)

    def test_constructor(self):
        name = reflection.get_callable_name(Class)
        self.assertEqual('.'.join((__name__, 'Class')), name)

    def test_callable_class(self):
        name = reflection.get_callable_name(CallableClass())
        self.assertEqual('.'.join((__name__, 'CallableClass')), name)

    def test_callable_class_call(self):
        name = reflection.get_callable_name(CallableClass().__call__)
        self.assertEqual(
            '.'.join((__name__, 'CallableClass', '__call__')), name
        )


class GetCallableNameTestExtended(test_base.BaseTestCase):
    # Tests items in http://legacy.python.org/dev/peps/pep-3155/

    class InnerCallableClass:
        def __call__(self):
            pass

    def test_inner_callable_class(self):
        obj = self.InnerCallableClass()
        name = reflection.get_callable_name(obj.__call__)
        expected_name = '.'.join(
            (
                __name__,
                'GetCallableNameTestExtended',
                'InnerCallableClass',
                '__call__',
            )
        )
        self.assertEqual(expected_name, name)

    def test_inner_callable_function(self):
        def a():
            def b():
                pass

            return b

        name = reflection.get_callable_name(a())
        expected_name = '.'.join(
            (
                __name__,
                'GetCallableNameTestExtended',
                'test_inner_callable_function',
                '<locals>',
                'a',
                '<locals>',
                'b',
            )
        )
        self.assertEqual(expected_name, name)

    def test_inner_class(self):
        obj = self.InnerCallableClass()
        name = reflection.get_callable_name(obj)
        expected_name = '.'.join(
            (__name__, 'GetCallableNameTestExtended', 'InnerCallableClass')
        )
        self.assertEqual(expected_name, name)


class GetCallableArgsTest(test_base.BaseTestCase):
    def test_mere_function(self):
        result = reflection.get_callable_args(mere_function)
        self.assertEqual(['a', 'b'], result)

    def test_function_with_defaults(self):
        result = reflection.get_callable_args(function_with_defs)
        self.assertEqual(['a', 'b', 'optional'], result)

    def test_required_only(self):
        result = reflection.get_callable_args(
            function_with_defs, required_only=True
        )
        self.assertEqual(['a', 'b'], result)

    def test_method(self):
        result = reflection.get_callable_args(Class.method)
        self.assertEqual(['self', 'c', 'd'], result)

    def test_instance_method(self):
        result = reflection.get_callable_args(Class().method)
        self.assertEqual(['c', 'd'], result)

    def test_class_method(self):
        result = reflection.get_callable_args(Class.class_method)
        self.assertEqual(['g', 'h'], result)

    def test_class_constructor(self):
        result = reflection.get_callable_args(ClassWithInit)
        self.assertEqual(['k', 'lll'], result)

    def test_class_with_call(self):
        result = reflection.get_callable_args(CallableClass())
        self.assertEqual(['i', 'j'], result)

    def test_decorators_work(self):
        @dummy_decorator
        def special_fun(x, y):
            pass

        result = reflection.get_callable_args(special_fun)
        self.assertEqual(['x', 'y'], result)


class AcceptsKwargsTest(test_base.BaseTestCase):
    def test_no_kwargs(self):
        self.assertEqual(False, reflection.accepts_kwargs(mere_function))

    def test_with_kwargs(self):
        self.assertEqual(True, reflection.accepts_kwargs(function_with_kwargs))


class GetClassNameTest(test_base.BaseTestCase):
    def test_std_exception(self):
        name = reflection.get_class_name(RuntimeError)
        self.assertEqual('RuntimeError', name)

    def test_class(self):
        name = reflection.get_class_name(Class)
        self.assertEqual('.'.join((__name__, 'Class')), name)

    def test_qualified_class(self):
        class QualifiedClass:
            pass

        name = reflection.get_class_name(QualifiedClass)
        self.assertEqual('.'.join((__name__, 'QualifiedClass')), name)

    def test_instance(self):
        name = reflection.get_class_name(Class())
        self.assertEqual('.'.join((__name__, 'Class')), name)

    def test_int(self):
        name = reflection.get_class_name(42)
        self.assertEqual('int', name)

    def test_class_method(self):
        name = reflection.get_class_name(Class.class_method)
        self.assertEqual(f'{__name__}.Class', name)
        # test with fully_qualified=False
        name = reflection.get_class_name(
            Class.class_method, fully_qualified=False
        )
        self.assertEqual('Class', name)

    def test_static_method(self):
        self.assertRaises(
            TypeError, reflection.get_class_name, Class.static_method
        )

    def test_unbound_method(self):
        self.assertRaises(TypeError, reflection.get_class_name, mere_function)

    def test_bound_method(self):
        c = Class()
        name = reflection.get_class_name(c.method)
        self.assertEqual(f'{__name__}.Class', name)
        # test with fully_qualified=False
        name = reflection.get_class_name(c.method, fully_qualified=False)
        self.assertEqual('Class', name)


class GetAllClassNamesTest(test_base.BaseTestCase):
    def test_std_class(self):
        names = list(reflection.get_all_class_names(RuntimeError))
        self.assertEqual(RUNTIME_ERROR_CLASSES, names)

    def test_std_class_up_to(self):
        names = list(
            reflection.get_all_class_names(RuntimeError, up_to=Exception)
        )
        self.assertEqual(RUNTIME_ERROR_CLASSES[:-2], names)



# Copyright 2011 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
File utilities.

.. versionadded:: 1.8
"""

import contextlib
import errno
import hashlib
import json
import os
import stat
import tempfile
import time
from typing import Any
from collections.abc import Callable, Generator

import yaml

from oslo_utils import excutils

_DEFAULT_MODE = stat.S_IRWXU | stat.S_IRWXG | stat.S_IRWXO


def ensure_tree(path: str, mode: int = _DEFAULT_MODE) -> None:
    """Create a directory (and any ancestor directories required)

    :param path: Directory to create
    :param mode: Directory creation permissions
    """
    try:
        os.makedirs(path, mode)
    except OSError as exc:
        if exc.errno == errno.EEXIST:
            if not os.path.isdir(path):
                raise
        else:
            raise


def delete_if_exists(
    path: str, remove: Callable[[str], Any] = os.unlink
) -> None:
    """Delete a file, but ignore file not found error.

    :param path: File to delete
    :param remove: Optional function to remove passed path
    """

    try:
        remove(path)
    except OSError as e:
        if e.errno != errno.ENOENT:
            raise


@contextlib.contextmanager
def remove_path_on_error(
    path: str, remove: Callable[[str], Any] = delete_if_exists
) -> Generator[None, None, None]:
    """Protect code that wants to operate on PATH atomically.
    Any exception will cause PATH to be removed.

    :param path: File to work with
    :param remove: Optional function to remove passed path
    """

    try:
        yield
    except Exception:
        with excutils.save_and_reraise_exception():
            remove(path)


def write_to_tempfile(
    content: bytes | bytearray,
    path: str | None = None,
    suffix: str = '',
    prefix: str = 'tmp',
) -> str:
    """Create a temporary file containing data.

    Create a temporary file containing specified content, with a specified
    filename suffix and prefix. The tempfile will be created in a default
    location, or in the directory `path`, if it is not None. `path` and its
    parent directories will be created if they don't exist.

    :param content: bytestring to write to the file
    :param path: same as parameter 'dir' for mkstemp
    :param suffix: same as parameter 'suffix' for mkstemp
    :param prefix: same as parameter 'prefix' for mkstemp

    For example: it can be used in database tests for creating
    configuration files.

    .. versionadded:: 1.9
    """
    if path:
        ensure_tree(path)

    fd, path = tempfile.mkstemp(suffix=suffix, dir=path, prefix=prefix)
    try:
        os.write(fd, content)
    finally:
        os.close(fd)
    return path


def compute_file_checksum(
    path: str, read_chunksize: int = 65536, algorithm: str = 'sha256'
) -> str:
    """Compute checksum of a file's contents.

    :param path: Path to the file
    :param read_chunksize: Maximum number of bytes to be read from the file
     at once. Default is 65536 bytes or 64KB
    :param algorithm: The hash algorithm name to use. For example, 'md5',
     'sha256', 'sha512' and so on. Default is 'sha256'. Refer to
     hashlib.algorithms_available for available algorithms
    :return: Hex digest string of the checksum

    .. versionadded:: 3.31.0
    """
    checksum = hashlib.new(algorithm)  # Raises appropriate exceptions.
    with open(path, 'rb') as f:
        for chunk in iter(lambda: f.read(read_chunksize), b''):
            checksum.update(chunk)
            # Release greenthread, if greenthreads are not used it is a noop.
            time.sleep(0)
    return checksum.hexdigest()


def last_bytes(path: str, num: int) -> tuple[bytes, int]:
    """Return num bytes from the end of the file and unread byte count.

    Returns a tuple containing some content from the file and the
    number of bytes that appear in the file before the point at which
    reading started. The content will be at most ``num`` bytes, taken
    from the end of the file. If the file is smaller than ``num``
    bytes the entire content of the file is returned.

    :param path: The file path to read
    :param num: The number of bytes to return

    :returns: (data, unread_bytes)

    """

    with open(path, 'rb') as fp:
        try:
            fp.seek(-num, os.SEEK_END)
        except OSError as e:
            # seek() fails with EINVAL when trying to go before the start of
            # the file. It means that num is larger than the file size, so
            # just go to the start.
            if e.errno == errno.EINVAL:
                fp.seek(0, os.SEEK_SET)
            else:
                raise
        unread_bytes = fp.tell()
        return (fp.read(), unread_bytes)


def is_json(file_path: str) -> bool:
    """Check if file is of json type or not.

    This function try to load the input file using json.loads()
    and return False if ValueError otherwise True.

    :param file_path: The file path to check

    :returns: bool

    """
    with open(file_path) as fh:
        data = fh.read()
    try:
        json.loads(data)
        return True
    except ValueError:
        return False


def is_yaml(file_path: str) -> bool:
    """Check if file is of yaml type or not.

    This function try to load the input file using yaml.safe_load()
    and return True if loadable. Because every json file can be loadable
    in yaml, so this function return False if file is loadable using
    json.loads() means it is json file.

    :param file_path: The file path to check

    :returns: bool

    """
    with open(file_path) as fh:
        data = fh.read()
        is_yaml = False
        try:
            json.loads(data)
        except ValueError:
            try:
                yaml.safe_load(data)
                is_yaml = True
            except yaml.scanner.ScannerError:
                pass
        return is_yaml



# Copyright (c) 2012 Intel Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import uuid

from oslotest import base as test_base

from oslo_utils import uuidutils


class UUIDUtilsTest(test_base.BaseTestCase):
    def test_generate_uuid(self):
        uuid_string = uuidutils.generate_uuid()
        self.assertIsInstance(uuid_string, str)
        self.assertEqual(len(uuid_string), 36)
        # make sure there are 4 dashes
        self.assertEqual(len(uuid_string.replace('-', '')), 32)

    def test_generate_uuid_dashed_false(self):
        uuid_string = uuidutils.generate_uuid(dashed=False)
        self.assertIsInstance(uuid_string, str)
        self.assertEqual(len(uuid_string), 32)
        self.assertNotIn('-', uuid_string)

    def test_is_uuid_like(self):
        self.assertTrue(uuidutils.is_uuid_like(str(uuid.uuid4())))
        self.assertTrue(
            uuidutils.is_uuid_like('{12345678-1234-5678-1234-567812345678}')
        )
        self.assertTrue(
            uuidutils.is_uuid_like('12345678123456781234567812345678')
        )
        self.assertTrue(
            uuidutils.is_uuid_like(
                'urn:uuid:12345678-1234-5678-1234-567812345678'
            )
        )
        self.assertTrue(
            uuidutils.is_uuid_like('urn:bbbaaaaa-aaaa-aaaa-aabb-bbbbbbbbbbbb')
        )
        self.assertTrue(
            uuidutils.is_uuid_like('uuid:bbbaaaaa-aaaa-aaaa-aabb-bbbbbbbbbbbb')
        )
        self.assertTrue(
            uuidutils.is_uuid_like(
                '{}---bbb---aaa--aaa--aaa-----aaa---aaa--bbb-bbb---bbb-bbb-bb-{}'
            )
        )

    def test_is_uuid_like_insensitive(self):
        self.assertTrue(uuidutils.is_uuid_like(str(uuid.uuid4()).upper()))

    def test_id_is_uuid_like(self):
        self.assertFalse(uuidutils.is_uuid_like(1234567))

    def test_name_is_uuid_like(self):
        self.assertFalse(uuidutils.is_uuid_like('zhongyueluo'))



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import hashlib
import hmac

from oslotest import base as test_base
import testscenarios

from oslo_utils import secretutils


def _gen_digest(text):
    return hmac.new(
        b'foo', text.encode('utf-8'), digestmod=hashlib.sha1
    ).digest()


class SecretUtilsTest(testscenarios.TestWithScenarios, test_base.BaseTestCase):
    scenarios = [
        ('binary', {'converter': _gen_digest}),
        ('unicode', {'converter': lambda text: text}),
    ]

    _test_data = b"Openstack forever"
    _md5_digest = hashlib.md5(_test_data).digest()

    def test_md5_with_data(self):
        digest = secretutils.md5(self._test_data).digest()
        self.assertEqual(digest, self._md5_digest)

        digest = secretutils.md5(
            self._test_data, usedforsecurity=True
        ).digest()
        self.assertEqual(digest, self._md5_digest)

        digest = secretutils.md5(
            self._test_data, usedforsecurity=False
        ).digest()
        self.assertEqual(digest, self._md5_digest)

    def test_md5_without_data(self):
        md5 = secretutils.md5()
        md5.update(self._test_data)
        digest = md5.digest()
        self.assertEqual(digest, self._md5_digest)

        md5 = secretutils.md5(usedforsecurity=True)
        md5.update(self._test_data)
        digest = md5.digest()
        self.assertEqual(digest, self._md5_digest)

        md5 = secretutils.md5(usedforsecurity=False)
        md5.update(self._test_data)
        digest = md5.digest()
        self.assertEqual(digest, self._md5_digest)

    def test_string_data_raises_type_error(self):
        self.assertRaises(TypeError, hashlib.md5, 'foo')
        self.assertRaises(TypeError, secretutils.md5, 'foo')
        self.assertRaises(
            TypeError, secretutils.md5, 'foo', usedforsecurity=True
        )
        self.assertRaises(
            TypeError, secretutils.md5, 'foo', usedforsecurity=False
        )

    def test_none_data_raises_type_error(self):
        self.assertRaises(TypeError, hashlib.md5, None)
        self.assertRaises(TypeError, secretutils.md5, None)
        self.assertRaises(
            TypeError, secretutils.md5, None, usedforsecurity=True
        )
        self.assertRaises(
            TypeError, secretutils.md5, None, usedforsecurity=False
        )

    def test_password_mksalt(self):
        self.assertRaises(ValueError, secretutils.crypt_mksalt, 'MD5')
        salt = secretutils.crypt_mksalt('SHA-256')
        self.assertEqual(3 + 16, len(salt))
        self.assertTrue(salt.startswith('$5$'))
        salt = secretutils.crypt_mksalt('SHA-512')
        self.assertEqual(3 + 16, len(salt))
        self.assertTrue(salt.startswith('$6$'))

    def test_password_crypt(self):
        self.assertEqual(
            '$5$mysalt$fcnMdhaFpUmeWtGOgVuImueZGL1v0Q1kUVbV2NbFOX4',
            secretutils.crypt_password('mytopsecret', '$5$mysalt$'),
        )
        self.assertEqual(
            '$6$mysalt$jTEJ24XtvcWmav/sTQb1tYqmk1kBQD/sxcMIxEPUcie'
            'J8L9AuCTWxYlxGz.XtIQYWspWkUXQz9zPIFTSKubP6.',
            secretutils.crypt_password('mytopsecret', '$6$mysalt$'),
        )



# Copyright 2015 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Test fixtures.

.. versionadded:: 1.3
"""

import threading

import fixtures

from oslo_utils import timeutils
from oslo_utils import uuidutils


class TimeFixture(fixtures.Fixture):
    """A fixture for overriding the time returned by timeutils.utcnow().

    :param override_time: datetime instance or list thereof. If not given,
                          defaults to the current UTC time.

    """

    def __init__(self, override_time=None):
        super().__init__()
        self._override_time = override_time

    def setUp(self):
        super().setUp()
        timeutils.set_time_override(self._override_time)
        self.addCleanup(timeutils.clear_time_override)

    def advance_time_delta(self, timedelta):
        """Advance overridden time using a datetime.timedelta."""
        timeutils.advance_time_delta(timedelta)

    def advance_time_seconds(self, seconds):
        """Advance overridden time by seconds."""
        timeutils.advance_time_seconds(seconds)


class _UUIDSentinels:
    """Registry of dynamically created, named, random UUID strings in regular
    (with hyphens) and similar to some keystone IDs (without hyphens) formats.

    An instance of this class will dynamically generate attributes as they are
    referenced, associating a random UUID string with each. Thereafter,
    referring to the same attribute will give the same UUID for the life of the
    instance. Plan accordingly.

    Usage::

        from oslo_utils.fixture import uuidsentinel as uuids
        from oslo_utils.fixture import keystoneidsentinel as keystids

        ...
        foo = uuids.foo
        do_a_thing(foo)
        # Referencing the same sentinel again gives the same value
        assert foo == uuids.foo
        # But a different one will be different
        assert foo != uuids.bar
        # Same approach is valid for keystoneidsentinel:
        data = create_some_data_structure(keystids.bar, var1, var2, var3)
        assert extract_bar(data) == keystids.bar
    """

    def __init__(self, is_dashed=True):
        self._sentinels = {}
        self._lock = threading.Lock()
        self.is_dashed = is_dashed

    def __getattr__(self, name):
        if name.startswith('_'):
            raise AttributeError('Sentinels must not start with _')
        with self._lock:
            if name not in self._sentinels:
                self._sentinels[name] = uuidutils.generate_uuid(
                    dashed=self.is_dashed
                )
        return self._sentinels[name]


# Singleton sentinel instance. Caveat emptor: using this multiple times in the
# same process (including across multiple modules) will result in the same
# values
uuidsentinel = _UUIDSentinels()
keystoneidsentinel = _UUIDSentinels(is_dashed=False)



# Copyright 2011 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
import logging
import time
from unittest import mock

import iso8601
from oslotest import base as test_base
from testtools import matchers

from oslo_utils import timeutils


def monotonic_iter(start=0, incr=0.05):
    while True:
        yield start
        start += incr


class TimeUtilsTest(test_base.BaseTestCase):
    def setUp(self):
        super().setUp()
        self.skynet_self_aware_time_str = '1997-08-29T06:14:00Z'
        self.skynet_self_aware_time_ms_str = '1997-08-29T06:14:00.000123Z'
        self.skynet_self_aware_time = datetime.datetime(1997, 8, 29, 6, 14, 0)
        self.skynet_self_aware_ms_time = datetime.datetime(
            1997, 8, 29, 6, 14, 0, 123
        )
        self.one_minute_before = datetime.datetime(1997, 8, 29, 6, 13, 0)
        self.one_minute_after = datetime.datetime(1997, 8, 29, 6, 15, 0)
        self.skynet_self_aware_time_perfect_str = '1997-08-29T06:14:00.000000'
        self.skynet_self_aware_time_perfect = datetime.datetime(
            1997, 8, 29, 6, 14, 0
        )
        self.addCleanup(timeutils.clear_time_override)

    def test_parse_isotime(self):
        expect = timeutils.parse_isotime(self.skynet_self_aware_time_str)
        skynet_self_aware_time_utc = self.skynet_self_aware_time.replace(
            tzinfo=iso8601.iso8601.UTC
        )
        self.assertEqual(skynet_self_aware_time_utc, expect)

    def test_parse_isotime_micro_second_precision(self):
        expect = timeutils.parse_isotime(self.skynet_self_aware_time_ms_str)
        skynet_self_aware_time_ms_utc = self.skynet_self_aware_ms_time.replace(
            tzinfo=iso8601.iso8601.UTC
        )
        self.assertEqual(skynet_self_aware_time_ms_utc, expect)

    def test_parse_strtime(self):
        perfect_time_format = self.skynet_self_aware_time_perfect_str
        expect = timeutils.parse_strtime(perfect_time_format)
        self.assertEqual(self.skynet_self_aware_time_perfect, expect)

    @mock.patch('datetime.datetime', wraps=datetime.datetime)
    def _test_is_older_than(self, fn, datetime_mock):
        datetime_mock.now.return_value = self.skynet_self_aware_time
        expect_true = timeutils.is_older_than(fn(self.one_minute_before), 59)
        self.assertTrue(expect_true)
        expect_false = timeutils.is_older_than(fn(self.one_minute_before), 60)
        self.assertFalse(expect_false)
        expect_false = timeutils.is_older_than(fn(self.one_minute_before), 61)
        self.assertFalse(expect_false)

    def test_is_older_than_datetime(self):
        self._test_is_older_than(lambda x: x)

    def test_is_older_than_aware(self):
        """Tests sending is_older_than an 'aware' datetime."""
        self._test_is_older_than(
            lambda x: x.replace(tzinfo=iso8601.iso8601.UTC)
        )

    def test_is_older_than_aware_no_utc(self):
        self._test_is_older_than(
            lambda x: x.replace(
                tzinfo=iso8601.iso8601.FixedOffset(1, 0, 'foo')
            ).replace(hour=7)
        )

    @mock.patch('datetime.datetime', wraps=datetime.datetime)
    def _test_is_newer_than(self, fn, datetime_mock):
        datetime_mock.now.return_value = self.skynet_self_aware_time
        expect_true = timeutils.is_newer_than(fn(self.one_minute_after), 59)
        self.assertTrue(expect_true)
        expect_false = timeutils.is_newer_than(fn(self.one_minute_after), 60)
        self.assertFalse(expect_false)
        expect_false = timeutils.is_newer_than(fn(self.one_minute_after), 61)
        self.assertFalse(expect_false)

    def test_is_newer_than_datetime(self):
        self._test_is_newer_than(lambda x: x)

    def test_is_newer_than_aware(self):
        """Tests sending is_newer_than an 'aware' datetime."""
        self._test_is_newer_than(
            lambda x: x.replace(tzinfo=iso8601.iso8601.UTC)
        )

    def test_is_newer_than_aware_no_utc(self):
        self._test_is_newer_than(
            lambda x: x.replace(
                tzinfo=iso8601.iso8601.FixedOffset(1, 0, 'foo')
            ).replace(hour=7)
        )

    def test_set_time_override_using_default(self):
        now = timeutils.utcnow_ts()

        # NOTE(kgriffs): Normally it's bad form to sleep in a unit test,
        # but this is the only way to test that set_time_override defaults
        # to setting the override to the current time.
        time.sleep(1)

        timeutils.set_time_override()
        overriden_now = timeutils.utcnow_ts()
        self.assertThat(now, matchers.LessThan(overriden_now))

    def test_utcnow_ts(self):
        skynet_self_aware_ts = 872835240
        skynet_dt = datetime.datetime.utcfromtimestamp(skynet_self_aware_ts)
        self.assertEqual(self.skynet_self_aware_time, skynet_dt)

        # NOTE(kgriffs): timeutils.utcnow_ts() uses time.time()
        # IFF time override is not set.
        with mock.patch('time.time') as time_mock:
            time_mock.return_value = skynet_self_aware_ts
            ts = timeutils.utcnow_ts()
            self.assertEqual(ts, skynet_self_aware_ts)

        timeutils.set_time_override(skynet_dt)
        ts = timeutils.utcnow_ts()
        self.assertEqual(ts, skynet_self_aware_ts)

    def test_utcnow(self):
        now = datetime.datetime.now()
        timeutils.set_time_override(now)
        self.assertEqual(timeutils.utcnow(), now)

        timeutils.clear_time_override()
        self.assertNotEqual(timeutils.utcnow(), now)

        self.assertTrue(timeutils.utcnow())

    def test_advance_time_delta(self):
        timeutils.set_time_override(self.one_minute_before)
        timeutils.advance_time_delta(datetime.timedelta(seconds=60))
        self.assertEqual(timeutils.utcnow(), self.skynet_self_aware_time)

    def test_advance_time_seconds(self):
        timeutils.set_time_override(self.one_minute_before)
        timeutils.advance_time_seconds(60)
        self.assertEqual(timeutils.utcnow(), self.skynet_self_aware_time)

    def test_marshall_time(self):
        now = timeutils.utcnow()
        binary = timeutils.marshall_now(now)
        backagain = timeutils.unmarshall_time(binary)
        self.assertEqual(now, backagain)

    def test_marshall_time_with_tz(self):
        now = timeutils.utcnow()
        now = now.replace(tzinfo=iso8601.iso8601.UTC)
        binary = timeutils.marshall_now(now)
        self.assertEqual("UTC", binary['tzname'])
        backagain = timeutils.unmarshall_time(binary)
        self.assertEqual(now, backagain)
        self.assertIsNotNone(backagain.tzinfo)
        self.assertEqual(now.utcoffset(), backagain.utcoffset())

    def test_unmarshall_time_leap_second(self):
        leap_dict = dict(
            day=30,
            month=6,
            year=2015,
            hour=23,
            minute=59,
            second=timeutils._MAX_DATETIME_SEC + 1,
            microsecond=0,
        )
        leap_time = timeutils.unmarshall_time(leap_dict)

        leap_dict.update(second=timeutils._MAX_DATETIME_SEC)
        expected = timeutils.unmarshall_time(leap_dict)

        self.assertEqual(expected, leap_time)

    def test_delta_seconds(self):
        before = timeutils.utcnow()
        after = before + datetime.timedelta(
            days=7, seconds=59, microseconds=123456
        )
        self.assertAlmostEqual(
            604859.123456, timeutils.delta_seconds(before, after)
        )

    def test_is_soon(self):
        expires = timeutils.utcnow() + datetime.timedelta(minutes=5)
        self.assertFalse(timeutils.is_soon(expires, 120))
        self.assertTrue(timeutils.is_soon(expires, 300))
        self.assertTrue(timeutils.is_soon(expires, 600))

        with mock.patch('datetime.datetime') as datetime_mock:
            datetime_mock.now.return_value = self.skynet_self_aware_time
            expires = timeutils.utcnow()
            self.assertTrue(timeutils.is_soon(expires, 0))


class TestIso8601Time(test_base.BaseTestCase):
    def _instaneous(self, timestamp, yr, mon, day, hr, minute, sec, micro):
        self.assertEqual(timestamp.year, yr)
        self.assertEqual(timestamp.month, mon)
        self.assertEqual(timestamp.day, day)
        self.assertEqual(timestamp.hour, hr)
        self.assertEqual(timestamp.minute, minute)
        self.assertEqual(timestamp.second, sec)
        self.assertEqual(timestamp.microsecond, micro)

    def _do_test(self, time_str, yr, mon, day, hr, minute, sec, micro, shift):
        DAY_SECONDS = 24 * 60 * 60
        timestamp = timeutils.parse_isotime(time_str)
        self._instaneous(timestamp, yr, mon, day, hr, minute, sec, micro)
        assert timestamp.tzinfo is not None
        offset = timestamp.tzinfo.utcoffset(None)
        assert offset is not None
        self.assertEqual(offset.seconds + offset.days * DAY_SECONDS, shift)

    def test_zulu(self):
        time_str = '2012-02-14T20:53:07Z'
        self._do_test(time_str, 2012, 2, 14, 20, 53, 7, 0, 0)

    def test_zulu_micros(self):
        time_str = '2012-02-14T20:53:07.123Z'
        self._do_test(time_str, 2012, 2, 14, 20, 53, 7, 123000, 0)

    def test_offset_east(self):
        time_str = '2012-02-14T20:53:07+04:30'
        offset = 4.5 * 60 * 60
        self._do_test(time_str, 2012, 2, 14, 20, 53, 7, 0, offset)

    def test_offset_east_micros(self):
        time_str = '2012-02-14T20:53:07.42+04:30'
        offset = 4.5 * 60 * 60
        self._do_test(time_str, 2012, 2, 14, 20, 53, 7, 420000, offset)

    def test_offset_west(self):
        time_str = '2012-02-14T20:53:07-05:30'
        offset = -5.5 * 60 * 60
        self._do_test(time_str, 2012, 2, 14, 20, 53, 7, 0, offset)

    def test_offset_west_micros(self):
        time_str = '2012-02-14T20:53:07.654321-05:30'
        offset = -5.5 * 60 * 60
        self._do_test(time_str, 2012, 2, 14, 20, 53, 7, 654321, offset)

    def test_compare(self):
        zulu = timeutils.parse_isotime('2012-02-14T20:53:07')
        east = timeutils.parse_isotime('2012-02-14T20:53:07-01:00')
        west = timeutils.parse_isotime('2012-02-14T20:53:07+01:00')
        self.assertTrue(east > west)
        self.assertTrue(east > zulu)
        self.assertTrue(zulu > west)

    def test_compare_micros(self):
        zulu = timeutils.parse_isotime('2012-02-14T20:53:07.6544')
        east = timeutils.parse_isotime('2012-02-14T19:53:07.654321-01:00')
        west = timeutils.parse_isotime('2012-02-14T21:53:07.655+01:00')
        self.assertTrue(east < west)
        self.assertTrue(east < zulu)
        self.assertTrue(zulu < west)

    def test_zulu_normalize(self):
        time_str = '2012-02-14T20:53:07Z'
        zulu = timeutils.parse_isotime(time_str)
        normed = timeutils.normalize_time(zulu)
        self._instaneous(normed, 2012, 2, 14, 20, 53, 7, 0)

    def test_east_normalize(self):
        time_str = '2012-02-14T20:53:07-07:00'
        east = timeutils.parse_isotime(time_str)
        normed = timeutils.normalize_time(east)
        self._instaneous(normed, 2012, 2, 15, 3, 53, 7, 0)

    def test_west_normalize(self):
        time_str = '2012-02-14T20:53:07+21:00'
        west = timeutils.parse_isotime(time_str)
        normed = timeutils.normalize_time(west)
        self._instaneous(normed, 2012, 2, 13, 23, 53, 7, 0)

    def test_normalize_aware_to_naive(self):
        dt = datetime.datetime(2011, 2, 14, 20, 53, 7)
        time_str = '2011-02-14T20:53:07+21:00'
        aware = timeutils.parse_isotime(time_str)
        naive = timeutils.normalize_time(aware)
        self.assertTrue(naive < dt)

    def test_normalize_zulu_aware_to_naive(self):
        dt = datetime.datetime(2011, 2, 14, 20, 53, 7)
        time_str = '2011-02-14T19:53:07Z'
        aware = timeutils.parse_isotime(time_str)
        naive = timeutils.normalize_time(aware)
        self.assertTrue(naive < dt)

    def test_normalize_naive(self):
        dt = datetime.datetime(2011, 2, 14, 20, 53, 7)
        dtn = datetime.datetime(2011, 2, 14, 19, 53, 7)
        naive = timeutils.normalize_time(dtn)
        self.assertTrue(naive < dt)


class TimeItTest(test_base.BaseTestCase):
    @mock.patch('time.sleep')
    @mock.patch('oslo_utils.timeutils.now')
    def test_timed(self, mock_now, mock_sleep):
        mock_now.side_effect = monotonic_iter(incr=0.1)
        fake_logger = mock.MagicMock(logging.getLogger(), autospec=True)

        @timeutils.time_it(fake_logger)
        def slow_function():
            time.sleep(0.1)

        slow_function()
        self.assertTrue(mock_now.called)
        self.assertTrue(mock_sleep.called)
        self.assertTrue(fake_logger.log.called)
        fake_logger.log.assert_called_with(logging.DEBUG, mock.ANY, mock.ANY)

    @mock.patch('time.sleep')
    @mock.patch('oslo_utils.timeutils.now')
    def test_no_timed_disabled(self, mock_now, mock_sleep):
        mock_now.side_effect = monotonic_iter(incr=0.1)
        fake_logger = mock.MagicMock(logging.getLogger(), autospec=True)

        @timeutils.time_it(fake_logger, enabled=False)
        def slow_function():
            time.sleep(0.1)

        slow_function()
        self.assertFalse(mock_now.called)
        self.assertFalse(fake_logger.log.called)

    @mock.patch('time.sleep')
    @mock.patch('oslo_utils.timeutils.now')
    def test_no_timed_to_fast(self, mock_now, mock_sleep):
        mock_now.side_effect = monotonic_iter(incr=0.1)
        fake_logger = mock.MagicMock(logging.getLogger(), autospec=True)

        @timeutils.time_it(fake_logger, min_duration=10)
        def fast_function():
            pass

        fast_function()
        self.assertFalse(fake_logger.log.called)

    @mock.patch('time.sleep')
    @mock.patch('oslo_utils.timeutils.now')
    def test_no_timed_exception(self, mock_now, mock_sleep):
        mock_now.side_effect = monotonic_iter(incr=0.1)
        fake_logger = mock.MagicMock(logging.getLogger(), autospec=True)

        @timeutils.time_it(fake_logger)
        def broken_function():
            raise OSError("Broken")

        self.assertRaises(IOError, broken_function)
        self.assertFalse(fake_logger.log.called)

    @mock.patch('time.sleep')
    @mock.patch('oslo_utils.timeutils.now')
    def test_timed_custom_message(self, mock_now, mock_sleep):
        mock_now.side_effect = monotonic_iter(incr=0.1)
        fake_logger = mock.MagicMock(logging.getLogger(), autospec=True)

        @timeutils.time_it(fake_logger, message="That took a long time")
        def slow_function():
            time.sleep(0.1)

        slow_function()
        self.assertTrue(mock_now.called)
        self.assertTrue(mock_sleep.called)
        self.assertTrue(fake_logger.log.called)
        fake_logger.log.assert_called_with(
            logging.DEBUG, "That took a long time", mock.ANY
        )

    @mock.patch('time.sleep')
    @mock.patch('oslo_utils.timeutils.now')
    def test_timed_custom_level(self, mock_now, mock_sleep):
        mock_now.side_effect = monotonic_iter(incr=0.1)
        fake_logger = mock.MagicMock(logging.getLogger(), autospec=True)

        @timeutils.time_it(fake_logger, log_level=logging.INFO)
        def slow_function():
            time.sleep(0.1)

        slow_function()
        self.assertTrue(mock_now.called)
        self.assertTrue(mock_sleep.called)
        self.assertTrue(fake_logger.log.called)
        fake_logger.log.assert_called_with(logging.INFO, mock.ANY, mock.ANY)


class StopWatchTest(test_base.BaseTestCase):
    def test_leftover_no_duration(self):
        watch = timeutils.StopWatch()
        watch.start()
        self.assertRaises(RuntimeError, watch.leftover)
        self.assertRaises(RuntimeError, watch.leftover, return_none=False)
        self.assertIsNone(watch.leftover(return_none=True))

    def test_no_states(self):
        watch = timeutils.StopWatch()
        self.assertRaises(RuntimeError, watch.stop)
        self.assertRaises(RuntimeError, watch.resume)

    def test_bad_expiry(self):
        self.assertRaises(ValueError, timeutils.StopWatch, -1)

    @mock.patch('oslo_utils.timeutils.now')
    def test_backwards(self, mock_now):
        mock_now.side_effect = [0, 0.5, -1.0, -1.0]
        watch = timeutils.StopWatch(0.1)
        watch.start()
        self.assertTrue(watch.expired())
        self.assertFalse(watch.expired())
        self.assertEqual(0.0, watch.elapsed())

    @mock.patch('oslo_utils.timeutils.now')
    def test_expiry(self, mock_now):
        mock_now.side_effect = monotonic_iter(incr=0.2)
        watch = timeutils.StopWatch(0.1)
        watch.start()
        self.assertTrue(watch.expired())

    @mock.patch('oslo_utils.timeutils.now')
    def test_not_expired(self, mock_now):
        mock_now.side_effect = monotonic_iter()
        watch = timeutils.StopWatch(0.1)
        watch.start()
        self.assertFalse(watch.expired())

    def test_has_started_stopped(self):
        watch = timeutils.StopWatch()
        self.assertFalse(watch.has_started())
        self.assertFalse(watch.has_stopped())
        watch.start()

        self.assertTrue(watch.has_started())
        self.assertFalse(watch.has_stopped())

        watch.stop()
        self.assertTrue(watch.has_stopped())
        self.assertFalse(watch.has_started())

    def test_no_expiry(self):
        watch = timeutils.StopWatch(0.1)
        self.assertRaises(RuntimeError, watch.expired)

    @mock.patch('oslo_utils.timeutils.now')
    def test_elapsed(self, mock_now):
        mock_now.side_effect = monotonic_iter(incr=0.2)
        watch = timeutils.StopWatch()
        watch.start()
        matcher = matchers.GreaterThan(0.19)
        self.assertThat(watch.elapsed(), matcher)

    def test_no_elapsed(self):
        watch = timeutils.StopWatch()
        self.assertRaises(RuntimeError, watch.elapsed)

    def test_no_leftover(self):
        watch = timeutils.StopWatch()
        self.assertRaises(RuntimeError, watch.leftover)
        watch = timeutils.StopWatch(1)
        self.assertRaises(RuntimeError, watch.leftover)

    @mock.patch('oslo_utils.timeutils.now')
    def test_pause_resume(self, mock_now):
        mock_now.side_effect = monotonic_iter()
        watch = timeutils.StopWatch()
        watch.start()
        watch.stop()
        elapsed = watch.elapsed()
        self.assertAlmostEqual(elapsed, watch.elapsed())
        watch.resume()
        self.assertNotEqual(elapsed, watch.elapsed())

    @mock.patch('oslo_utils.timeutils.now')
    def test_context_manager(self, mock_now):
        mock_now.side_effect = monotonic_iter()
        with timeutils.StopWatch() as watch:
            pass
        matcher = matchers.GreaterThan(0.04)
        self.assertThat(watch.elapsed(), matcher)

    @mock.patch('oslo_utils.timeutils.now')
    def test_context_manager_splits(self, mock_now):
        mock_now.side_effect = monotonic_iter()
        with timeutils.StopWatch() as watch:
            time.sleep(0.01)
            watch.split()
        self.assertRaises(RuntimeError, watch.split)
        self.assertEqual(1, len(watch.splits))

    def test_splits_stopped(self):
        watch = timeutils.StopWatch()
        watch.start()
        watch.split()
        watch.stop()
        self.assertRaises(RuntimeError, watch.split)

    def test_splits_never_started(self):
        watch = timeutils.StopWatch()
        self.assertRaises(RuntimeError, watch.split)

    @mock.patch('oslo_utils.timeutils.now')
    def test_splits(self, mock_now):
        mock_now.side_effect = monotonic_iter()

        watch = timeutils.StopWatch()
        watch.start()
        self.assertEqual(0, len(watch.splits))

        watch.split()
        self.assertEqual(1, len(watch.splits))
        self.assertEqual(watch.splits[0].elapsed, watch.splits[0].length)

        watch.split()
        splits = watch.splits
        self.assertEqual(2, len(splits))
        self.assertNotEqual(splits[0].elapsed, splits[1].elapsed)
        self.assertEqual(
            splits[1].length, splits[1].elapsed - splits[0].elapsed
        )

        watch.stop()
        self.assertEqual(2, len(watch.splits))

        watch.start()
        self.assertEqual(0, len(watch.splits))

    @mock.patch('oslo_utils.timeutils.now')
    def test_elapsed_maximum(self, mock_now):
        mock_now.side_effect = [0, 1] + ([11] * 4)

        watch = timeutils.StopWatch()
        watch.start()
        self.assertEqual(1, watch.elapsed())

        self.assertEqual(11, watch.elapsed())
        self.assertEqual(1, watch.elapsed(maximum=1))

        watch.stop()
        self.assertEqual(11, watch.elapsed())
        self.assertEqual(11, watch.elapsed())
        self.assertEqual(0, watch.elapsed(maximum=-1))



# Copyright (c) 2013 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Helpers for comparing version strings.

.. versionadded:: 1.6
"""

import functools
import operator
import re

import packaging.version

from oslo_utils._i18n import _


def is_compatible(
    requested_version: str, current_version: str, same_major: bool = True
) -> bool:
    """Determine whether `requested_version` is satisfied by
    `current_version`; in other words, `current_version` is >=
    `requested_version`.

    :param requested_version: version to check for compatibility
    :param current_version: version to check against
    :param same_major: if True, the major version must be identical between
        `requested_version` and `current_version`. This is used when a
        major-version difference indicates incompatibility between the two
        versions. Since this is the common-case in practice, the default is
        True.
    :returns: True if compatible, False if not
    """
    requested = packaging.version.Version(requested_version)
    current = packaging.version.Version(current_version)

    if same_major:
        if requested.major != current.major:
            return False

    return bool(current >= requested)


def convert_version_to_int(version: str | tuple[int, ...]) -> int:
    """Convert a version to an integer.

    *version* must be a string with dots or a tuple of integers.

    .. versionadded:: 2.0
    """
    try:
        if isinstance(version, str):
            version = convert_version_to_tuple(version)
        if isinstance(version, tuple):
            return functools.reduce(lambda x, y: (x * 1000) + y, version)
    except Exception as ex:
        msg = _("Version %s is invalid.") % version
        raise ValueError(msg) from ex


def convert_version_to_str(version_int: int) -> str:
    """Convert a version integer to a string with dots.

    .. versionadded:: 2.0
    """
    version_numbers: list[int] = []
    factor = 1000
    while version_int != 0:
        version_number = version_int - (version_int // factor * factor)
        version_numbers.insert(0, version_number)
        version_int = version_int // factor

    return '.'.join(map(str, version_numbers))


def convert_version_to_tuple(version_str: str) -> tuple[int, ...]:
    """Convert a version string with dots to a tuple.

    .. versionadded:: 2.0
    """
    version_str = re.sub(r'(\d+)(a|alpha|b|beta|rc)\d+$', '\\1', version_str)
    return tuple(int(part) for part in version_str.split('.'))


class VersionPredicate:
    """Parse version predicate and check version requirements

    This is based on the implementation of distutils.VersionPredicate

    .. versionadded:: 7.4
    """

    _PREDICATE_MATCH = re.compile(r"^\s*(<=|>=|<|>|!=|==)\s*([^\s]+)\s*$")
    _COMP_MAP = {
        "<": operator.lt,
        "<=": operator.le,
        "==": operator.eq,
        ">": operator.gt,
        ">=": operator.ge,
        "!=": operator.ne,
    }

    def __init__(self, predicate_str: str) -> None:
        self.pred = [
            self._parse_predicate(pred) for pred in predicate_str.split(',')
        ]

    def _parse_predicate(
        self, pred: str
    ) -> tuple[str, packaging.version.Version]:
        res = self._PREDICATE_MATCH.match(pred)
        if not res:
            raise ValueError(f"bad package restriction syntax: {pred}")
        cond, ver_str = res.groups()
        return (cond, packaging.version.Version(ver_str))

    def satisfied_by(self, version_str: str) -> bool:
        version = packaging.version.Version(version_str)
        for cond, ver in self.pred:
            if not self._COMP_MAP[cond](version, ver):
                return False
        return True



# Copyright 2012 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from collections import namedtuple
import contextlib
import io
import socket
from unittest import mock

import netaddr
from oslotest import base as test_base

from oslo_utils import netutils


class NetworkUtilsTest(test_base.BaseTestCase):
    def test_no_host(self):
        result = netutils.urlsplit('http://')
        self.assertEqual('', result.netloc)
        self.assertIsNone(result.port)
        self.assertIsNone(result.hostname)
        self.assertEqual('http', result.scheme)

    def test_parse_host_port(self):
        self.assertEqual(
            ('server01', 80), netutils.parse_host_port('server01:80')
        )
        self.assertEqual(
            ('server01', None), netutils.parse_host_port('server01')
        )
        self.assertEqual(
            ('server01', 1234),
            netutils.parse_host_port('server01', default_port=1234),
        )
        self.assertEqual(('::1', 80), netutils.parse_host_port('[::1]:80'))
        self.assertEqual(('::1', None), netutils.parse_host_port('[::1]'))
        self.assertEqual(
            ('::1', 1234), netutils.parse_host_port('[::1]', default_port=1234)
        )
        self.assertEqual(
            ('2001:db8:85a3::8a2e:370:7334', 1234),
            netutils.parse_host_port(
                '2001:db8:85a3::8a2e:370:7334', default_port=1234
            ),
        )

    def test_urlsplit(self):
        result = netutils.urlsplit('rpc://myhost?someparam#somefragment')
        self.assertEqual(result.scheme, 'rpc')
        self.assertEqual(result.netloc, 'myhost')
        self.assertEqual(result.path, '')
        self.assertEqual(result.query, 'someparam')
        self.assertEqual(result.fragment, 'somefragment')

        result = netutils.urlsplit(
            'rpc://myhost/mypath?someparam#somefragment', allow_fragments=False
        )
        self.assertEqual(result.scheme, 'rpc')
        self.assertEqual(result.netloc, 'myhost')
        self.assertEqual(result.path, '/mypath')
        self.assertEqual(result.query, 'someparam#somefragment')
        self.assertEqual(result.fragment, '')

        result = netutils.urlsplit(
            'rpc://user:pass@myhost/mypath?someparam#somefragment',
            allow_fragments=False,
        )
        self.assertEqual(result.scheme, 'rpc')
        self.assertEqual(result.netloc, 'user:pass@myhost')
        self.assertEqual(result.path, '/mypath')
        self.assertEqual(result.query, 'someparam#somefragment')
        self.assertEqual(result.fragment, '')

    def test_urlsplit_ipv6(self):
        ipv6_url = 'http://[::1]:443/v2.0/'
        result = netutils.urlsplit(ipv6_url)
        self.assertEqual(result.scheme, 'http')
        self.assertEqual(result.netloc, '[::1]:443')
        self.assertEqual(result.path, '/v2.0/')
        self.assertEqual(result.hostname, '::1')
        self.assertEqual(result.port, 443)

        ipv6_url = 'http://user:pass@[::1]/v2.0/'
        result = netutils.urlsplit(ipv6_url)
        self.assertEqual(result.scheme, 'http')
        self.assertEqual(result.netloc, 'user:pass@[::1]')
        self.assertEqual(result.path, '/v2.0/')
        self.assertEqual(result.hostname, '::1')
        self.assertIsNone(result.port)

        ipv6_url = 'https://[2001:db8:85a3::8a2e:370:7334]:1234/v2.0/xy?ab#12'
        result = netutils.urlsplit(ipv6_url)
        self.assertEqual(result.scheme, 'https')
        self.assertEqual(result.netloc, '[2001:db8:85a3::8a2e:370:7334]:1234')
        self.assertEqual(result.path, '/v2.0/xy')
        self.assertEqual(result.hostname, '2001:db8:85a3::8a2e:370:7334')
        self.assertEqual(result.port, 1234)
        self.assertEqual(result.query, 'ab')
        self.assertEqual(result.fragment, '12')

    def test_urlsplit_params(self):
        test_url = "http://localhost/?a=b&c=d"
        result = netutils.urlsplit(test_url)
        self.assertEqual({'a': 'b', 'c': 'd'}, result.params())
        self.assertEqual({'a': 'b', 'c': 'd'}, result.params(collapse=False))

        test_url = "http://localhost/?a=b&a=c&a=d"
        result = netutils.urlsplit(test_url)
        self.assertEqual({'a': 'd'}, result.params())
        self.assertEqual({'a': ['b', 'c', 'd']}, result.params(collapse=False))

        test_url = "http://localhost"
        result = netutils.urlsplit(test_url)
        self.assertEqual({}, result.params())

        test_url = "http://localhost?"
        result = netutils.urlsplit(test_url)
        self.assertEqual({}, result.params())

    def test_set_tcp_keepalive(self):
        mock_sock = mock.Mock()
        netutils.set_tcp_keepalive(mock_sock, True, 100, 10, 5)
        calls = [
            mock.call.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, True),
        ]
        if hasattr(socket, 'TCP_KEEPIDLE'):
            calls += [
                mock.call.setsockopt(
                    socket.IPPROTO_TCP, socket.TCP_KEEPIDLE, 100
                )
            ]
        if hasattr(socket, 'TCP_KEEPINTVL'):
            calls += [
                mock.call.setsockopt(
                    socket.IPPROTO_TCP, socket.TCP_KEEPINTVL, 10
                ),
            ]
        if hasattr(socket, 'TCP_KEEPCNT'):
            calls += [
                mock.call.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPCNT, 5)
            ]
        mock_sock.assert_has_calls(calls)

        mock_sock.reset_mock()
        netutils.set_tcp_keepalive(mock_sock, False)
        self.assertEqual(1, len(mock_sock.mock_calls))

    @mock.patch.object(netutils, 'LOG', autospec=True)
    def test_is_valid_ipv4(self, mock_log):
        self.assertTrue(netutils.is_valid_ipv4('42.42.42.42'))

        self.assertFalse(netutils.is_valid_ipv4('-1.11.11.11'))

        self.assertFalse(netutils.is_valid_ipv4(''))

        self.assertFalse(netutils.is_valid_ipv4('10'))
        self.assertFalse(netutils.is_valid_ipv4('10.10'))
        self.assertFalse(netutils.is_valid_ipv4('10.10.10'))
        self.assertTrue(netutils.is_valid_ipv4('10.10.10.10'))
        mock_log.warning.assert_not_called()
        mock_log.reset_mock()

        self.assertFalse(netutils.is_valid_ipv4('10', strict=True))
        self.assertFalse(netutils.is_valid_ipv4('10.10', strict=True))
        self.assertFalse(netutils.is_valid_ipv4('10.10.10', strict=True))
        mock_log.warning.assert_not_called()
        mock_log.reset_mock()
        self.assertTrue(netutils.is_valid_ipv4('10', strict=False))
        self.assertTrue(netutils.is_valid_ipv4('10.10', strict=False))
        self.assertTrue(netutils.is_valid_ipv4('10.10.10', strict=False))
        mock_log.warning.assert_not_called()
        mock_log.reset_mock()

    def test_is_valid_ipv6(self):
        self.assertTrue(netutils.is_valid_ipv6('::1'))

        self.assertTrue(netutils.is_valid_ipv6('fe80::1%eth0'))

        self.assertFalse(netutils.is_valid_ip('fe%80::1%eth0'))

        self.assertFalse(
            netutils.is_valid_ipv6('1fff::a88:85a3::172.31.128.1')
        )

        self.assertFalse(netutils.is_valid_ipv6(''))

    def test_get_noscope_ipv6(self):
        self.assertEqual(
            '2001:db8::ff00:42:8329',
            netutils.get_noscope_ipv6('2001:db8::ff00:42:8329%1'),
        )
        self.assertEqual(
            'ff02::5678', netutils.get_noscope_ipv6('ff02::5678%eth0')
        )
        self.assertEqual('fe80::1', netutils.get_noscope_ipv6('fe80::1%eth0'))
        self.assertEqual('::1', netutils.get_noscope_ipv6('::1%eth0'))
        self.assertEqual('::1', netutils.get_noscope_ipv6('::1'))
        self.assertRaises(ValueError, netutils.get_noscope_ipv6, '::132:::')

    def test_escape_ipv6(self):
        self.assertEqual('[1234::1234]', netutils.escape_ipv6('1234::1234'))
        self.assertEqual('127.0.0.1', netutils.escape_ipv6('127.0.0.1'))

    def test_is_valid_ip(self):
        self.assertTrue(netutils.is_valid_ip('127.0.0.1'))

        self.assertTrue(netutils.is_valid_ip('2001:db8::ff00:42:8329'))

        self.assertTrue(netutils.is_valid_ip('fe80::1%eth0'))

        self.assertFalse(netutils.is_valid_ip('256.0.0.0'))

        self.assertFalse(netutils.is_valid_ip('::1.2.3.'))

        self.assertFalse(netutils.is_valid_ip(''))

        self.assertFalse(netutils.is_valid_ip(None))

    def test_is_valid_mac(self):
        self.assertTrue(netutils.is_valid_mac("52:54:00:cf:2d:31"))
        self.assertFalse(netutils.is_valid_mac("127.0.0.1"))
        self.assertFalse(netutils.is_valid_mac("not:a:mac:address"))
        self.assertFalse(netutils.is_valid_mac("52-54-00-cf-2d-31"))
        self.assertFalse(netutils.is_valid_mac("aa bb cc dd ee ff"))
        self.assertTrue(netutils.is_valid_mac("AA:BB:CC:DD:EE:FF"))
        self.assertFalse(netutils.is_valid_mac("AA BB CC DD EE FF"))
        self.assertFalse(netutils.is_valid_mac("AA-BB-CC-DD-EE-FF"))

    def test_is_valid_cidr(self):
        self.assertTrue(netutils.is_valid_cidr('10.0.0.0/24'))
        self.assertTrue(netutils.is_valid_cidr('10.0.0.1/32'))
        self.assertTrue(netutils.is_valid_cidr('0.0.0.0/0'))
        self.assertTrue(netutils.is_valid_cidr('2600::/64'))
        self.assertTrue(
            netutils.is_valid_cidr(
                '0000:0000:0000:0000:0000:0000:0000:0001/32'
            )
        )

        self.assertFalse(netutils.is_valid_cidr('10.0.0.1'))
        self.assertFalse(netutils.is_valid_cidr('10.0.0.1/33'))
        self.assertFalse(netutils.is_valid_cidr(10))  # type: ignore

    def test_is_valid_ipv6_cidr(self):
        self.assertTrue(netutils.is_valid_ipv6_cidr("2600::/64"))
        self.assertTrue(
            netutils.is_valid_ipv6_cidr(
                "abcd:ef01:2345:6789:abcd:ef01:192.168.254.254/48"
            )
        )
        self.assertTrue(
            netutils.is_valid_ipv6_cidr(
                "0000:0000:0000:0000:0000:0000:0000:0001/32"
            )
        )
        self.assertTrue(
            netutils.is_valid_ipv6_cidr(
                "0000:0000:0000:0000:0000:0000:0000:0001"
            )
        )
        self.assertFalse(netutils.is_valid_ipv6_cidr("foo"))
        self.assertFalse(netutils.is_valid_ipv6_cidr("127.0.0.1"))

    def test_valid_port(self):
        valid_inputs = [
            0,
            '0',
            1,
            '1',
            2,
            '3',
            '5',
            8,
            13,
            21,
            '80',
            '3246',
            '65535',
        ]
        for input_str in valid_inputs:
            self.assertTrue(netutils.is_valid_port(input_str))  # type: ignore

    def test_valid_port_fail(self):
        invalid_inputs = [
            '-32768',
            '65536',
            528491,
            '528491',
            '528.491',
            'thirty-seven',
            None,
        ]
        for input_str in invalid_inputs:
            self.assertFalse(netutils.is_valid_port(input_str))  # type: ignore

    def test_get_my_ipv4(self):
        mock_sock = mock.Mock()
        mock_sock.getsockname.return_value = ['1.2.3.4', '']
        sock_attrs = {'return_value.__enter__.return_value': mock_sock}
        with mock.patch('socket.socket', **sock_attrs):
            addr = netutils.get_my_ipv4()
        self.assertEqual(addr, '1.2.3.4')

    def test_get_my_ipv4_disabled(self):
        with (
            mock.patch('socket.socket', side_effect=OSError()),
            mock.patch('builtins.open', side_effect=FileNotFoundError()),
        ):
            addr = netutils.get_my_ipv4()
        self.assertEqual(addr, '127.0.0.1')

    def test_get_my_ipv6(self):
        mock_sock = mock.Mock()
        mock_sock.getsockname.return_value = ['2001:db8::2', '', '', '']
        sock_attrs = {'return_value.__enter__.return_value': mock_sock}
        with mock.patch('socket.socket', **sock_attrs):
            addr = netutils.get_my_ipv6()
        self.assertEqual(addr, '2001:db8::2')

    def test_get_my_ipv6_disabled(self):
        with (
            mock.patch('socket.socket', side_effect=OSError()),
            mock.patch('builtins.open', side_effect=FileNotFoundError()),
        ):
            addr = netutils.get_my_ipv6()
        self.assertEqual(addr, '::1')

    def test_is_int_in_range(self):
        valid_inputs = [
            (1, -100, 100),
            ('1', -100, 100),
            (100, -100, 100),
            ('100', -100, 100),
            (-100, -100, 100),
            ('-100', -100, 100),
        ]
        for input_value in valid_inputs:
            self.assertTrue(netutils._is_int_in_range(*input_value))  # type: ignore

    def test_is_int_not_in_range(self):
        invalid_inputs = [
            (None, 1, 100),
            ('ten', 1, 100),
            (-1, 0, 255),
            ('None', 1, 100),
        ]
        for input_value in invalid_inputs:
            self.assertFalse(netutils._is_int_in_range(*input_value))  # type: ignore

    def test_valid_icmp_type(self):
        valid_inputs = [1, '1', 0, '0', 255, '255']
        for input_value in valid_inputs:
            self.assertTrue(netutils.is_valid_icmp_type(input_value))  # type: ignore

    def test_invalid_icmp_type(self):
        invalid_inputs = [-1, '-1', 256, '256', None, 'None', 'five']
        for input_value in invalid_inputs:
            self.assertFalse(netutils.is_valid_icmp_type(input_value))  # type: ignore

    def test_valid_icmp_code(self):
        valid_inputs = [1, '1', 0, '0', 255, '255', None]
        for input_value in valid_inputs:
            self.assertTrue(netutils.is_valid_icmp_code(input_value))  # type: ignore

    def test_invalid_icmp_code(self):
        invalid_inputs = [-1, '-1', 256, '256', 'None', 'zero']
        for input_value in invalid_inputs:
            self.assertFalse(netutils.is_valid_icmp_code(input_value))  # type: ignore

    @mock.patch('socket.socket')
    @mock.patch('oslo_utils.netutils._get_my_ipv4_address')
    def test_get_my_ipv4_socket_error(self, ip, mock_socket):
        mock_socket.side_effect = socket.error
        ip.return_value = '1.2.3.4'
        addr = netutils.get_my_ipv4()
        self.assertEqual(addr, '1.2.3.4')

    @mock.patch('socket.socket')
    @mock.patch('oslo_utils.netutils._get_my_ipv6_address')
    def test_get_my_ipv6_socket_error(self, ip, mock_socket):
        mock_socket.side_effect = socket.error
        ip.return_value = '2001:db8::2'
        addr = netutils.get_my_ipv6()
        self.assertEqual(addr, '2001:db8::2')

    @mock.patch('builtins.open')
    @mock.patch('psutil.net_if_addrs')
    def test_get_my_ipv4_address_with_default_route(
        self, mock_ifaddrs, mock_open
    ):
        mock_open.return_value = io.StringIO(
            """Iface	Destination	Gateway 	Flags	RefCnt	Use	Metric	Mask		MTU	Window	IRTT
eth0	00000000	01cc12ac	0003	0	0	600	00000000	0	0	0
eth0	00cc12ac	00000000	0001	0	0	600	00FFFFFF	0	0	0
eth1	00cd12ac	00000000	0001	0	0	600	00FFFFFF	0	0	0"""
        )  # noqa : E501

        addr = namedtuple('addr', ['family', 'address'])
        mock_ifaddrs.return_value = {
            'eth0': [
                addr(family=socket.AF_INET, address='172.18.204.2'),
                addr(family=socket.AF_INET6, address='2001:db8::2'),
            ],
            'eth1': [
                addr(family=socket.AF_INET, address='172.18.205.2'),
                addr(family=socket.AF_INET6, address='2001:db8::1000::2'),
            ],
        }
        self.assertEqual('172.18.204.2', netutils._get_my_ipv4_address())
        mock_open.assert_called_once_with('/proc/net/route')

    @mock.patch('builtins.open')
    @mock.patch('psutil.net_if_addrs')
    def test_get_my_ipv6_address_with_default_route(
        self, mock_ifaddrs, mock_open
    ):
        mock_open.return_value = io.StringIO(
            """00000000000000000000000000000000 00 00000000000000000000000000000000 00 20010db8000000000000000000000001 00000000 00000000 00000000 08000000 eth0
20010db8000000000000000000000000 31 00000000000000000000000000000000 00 00000000000000000000000000000000 00000000 00000000 00000000 08000000 eth0
20010db8100000000000000000000000 31 00000000000000000000000000000000 00 00000000000000000000000000000000 00000000 00000000 00000000 08000000 eth1"""
        )  # noqa: E501

        addr = namedtuple('addr', ['family', 'address'])
        mock_ifaddrs.return_value = {
            'eth0': [
                addr(family=socket.AF_INET, address='172.18.204.2'),
                addr(family=socket.AF_INET6, address='2001:db8::2'),
            ],
            'eth1': [
                addr(family=socket.AF_INET, address='172.18.205.2'),
                addr(family=socket.AF_INET6, address='2001:db8::1000::2'),
            ],
        }
        self.assertEqual('2001:db8::2', netutils._get_my_ipv6_address())
        mock_open.assert_called_once_with('/proc/net/ipv6_route')

    @mock.patch('builtins.open')
    @mock.patch('psutil.net_if_addrs')
    def test_get_my_ipv4_address_without_default_route(
        self, mock_ifaddrs, mock_open
    ):
        mock_open.return_value = io.StringIO(
            """Iface	Destination	Gateway 	Flags	RefCnt	Use	Metric	Mask		MTU	Window	IRTT
eth0	00cc12ac	00000000	0001	0	0	600	00FFFFFF	0	0	0
eth1	00cd12ac	00000000	0001	0	0	600	00FFFFFF	0	0	0"""
        )  # noqa : E501

        self.assertEqual('127.0.0.1', netutils._get_my_ipv4_address())
        mock_open.assert_called_once_with('/proc/net/route')
        mock_ifaddrs.assert_not_called()

    @mock.patch('builtins.open')
    @mock.patch('psutil.net_if_addrs')
    def test_get_my_ipv6_address_without_default_route(
        self, mock_ifaddrs, mock_open
    ):
        mock_open.return_value = io.StringIO(
            """20010db8000000000000000000000000 31 00000000000000000000000000000000 00 00000000000000000000000000000000 00000000 00000000 00000000 08000000 eth0
20010db8100000000000000000000000 31 00000000000000000000000000000000 00 00000000000000000000000000000000 00000000 00000000 00000000 08000000 eth1"""
        )  # noqa: E501

        self.assertEqual('::1', netutils._get_my_ipv6_address())
        mock_open.assert_called_once_with('/proc/net/ipv6_route')
        mock_ifaddrs.assert_not_called()


class IPv6byEUI64TestCase(test_base.BaseTestCase):
    """Unit tests to generate IPv6 by EUI-64 operations."""

    def test_generate_IPv6_by_EUI64(self):
        addr = netutils.get_ipv6_addr_by_EUI64(
            '2001:db8::', '00:16:3e:33:44:55'
        )
        self.assertEqual('2001:db8::216:3eff:fe33:4455', addr.format())

    def test_generate_IPv6_with_IPv4_prefix(self):
        ipv4_prefix = '10.0.8'
        mac = '00:16:3e:33:44:55'
        self.assertRaises(
            ValueError,
            lambda: netutils.get_ipv6_addr_by_EUI64(ipv4_prefix, mac),
        )

    def test_generate_IPv6_with_bad_mac(self):
        bad_mac = '00:16:3e:33:44:5Z'
        prefix = '2001:db8::'
        self.assertRaises(
            ValueError,
            lambda: netutils.get_ipv6_addr_by_EUI64(prefix, bad_mac),
        )

    def test_generate_IPv6_with_bad_prefix(self):
        mac = '00:16:3e:33:44:55'
        bad_prefix = 'bb'
        self.assertRaises(
            ValueError,
            lambda: netutils.get_ipv6_addr_by_EUI64(bad_prefix, mac),
        )

    def test_generate_IPv6_with_error_prefix_type(self):
        mac = '00:16:3e:33:44:55'
        prefix = 123
        self.assertRaises(
            TypeError,
            lambda: netutils.get_ipv6_addr_by_EUI64(prefix, mac),  # type: ignore
        )

    def test_generate_IPv6_with_empty_prefix(self):
        mac = '00:16:3e:33:44:55'
        prefix = ''
        self.assertRaises(
            ValueError, lambda: netutils.get_ipv6_addr_by_EUI64(prefix, mac)
        )


class MACbyIPv6TestCase(test_base.BaseTestCase):
    """Unit tests to extract MAC from IPv6."""

    def test_reverse_generate_IPv6_by_EUI64(self):
        self.assertEqual(
            netaddr.EUI('00:16:3e:33:44:55'),
            netutils.get_mac_addr_by_ipv6(
                netaddr.IPAddress('2001:db8::216:3eff:fe33:4455')
            ),
        )

    def test_random_qemu_mac(self):
        self.assertEqual(
            netaddr.EUI('52:54:00:42:02:19'),
            netutils.get_mac_addr_by_ipv6(
                netaddr.IPAddress('fe80::5054:ff:fe42:219')
            ),
        )

    def test_local(self):
        self.assertEqual(
            netaddr.EUI('02:00:00:00:00:00'),
            netutils.get_mac_addr_by_ipv6(
                netaddr.IPAddress('fe80::ff:fe00:0')
            ),
        )

    def test_universal(self):
        self.assertEqual(
            netaddr.EUI('00:00:00:00:00:00'),
            netutils.get_mac_addr_by_ipv6(
                netaddr.IPAddress('fe80::200:ff:fe00:0')
            ),
        )


@contextlib.contextmanager
def mock_file_content(content):
    # Allows StringIO to act like a context manager-enabled file.
    yield io.StringIO(content)


class TestIsIPv6Enabled(test_base.BaseTestCase):
    def setUp(self):
        super().setUp()

        def reset_detection_flag():
            netutils._IS_IPV6_ENABLED = None

        reset_detection_flag()
        self.addCleanup(reset_detection_flag)

    @mock.patch('os.path.exists', return_value=True)
    @mock.patch('builtins.open', return_value=mock_file_content('0'))
    def test_enabled(self, mock_open, exists):
        enabled = netutils.is_ipv6_enabled()
        self.assertTrue(enabled)

    @mock.patch('os.path.exists', return_value=True)
    @mock.patch('builtins.open', return_value=mock_file_content('1'))
    def test_disabled(self, mock_open, exists):
        enabled = netutils.is_ipv6_enabled()
        self.assertFalse(enabled)

    @mock.patch('os.path.exists', return_value=False)
    @mock.patch('builtins.open', side_effect=AssertionError('should not read'))
    def test_disabled_non_exists(self, mock_open, exists):
        enabled = netutils.is_ipv6_enabled()
        self.assertFalse(enabled)

    @mock.patch('os.path.exists', return_value=True)
    def test_memoize_enabled(self, exists):
        # Reset the flag to appear that we haven't looked for it yet.
        netutils._IS_IPV6_ENABLED = None
        with mock.patch(
            'builtins.open', return_value=mock_file_content('0')
        ) as mock_open:
            enabled = netutils.is_ipv6_enabled()
            self.assertTrue(mock_open.called)
            self.assertTrue(netutils._IS_IPV6_ENABLED)
            self.assertTrue(enabled)
        # The second call should not use open again
        with mock.patch(
            'builtins.open', side_effect=AssertionError('should not be called')
        ):
            enabled = netutils.is_ipv6_enabled()
            self.assertTrue(enabled)

    @mock.patch('os.path.exists', return_value=True)
    def test_memoize_disabled(self, exists):
        # Reset the flag to appear that we haven't looked for it yet.
        netutils._IS_IPV6_ENABLED = None
        with mock.patch('builtins.open', return_value=mock_file_content('1')):
            enabled = netutils.is_ipv6_enabled()
            self.assertFalse(enabled)
        # The second call should not use open again
        with mock.patch(
            'builtins.open', side_effect=AssertionError('should not be called')
        ):
            enabled = netutils.is_ipv6_enabled()
            self.assertFalse(enabled)

    @mock.patch('os.path.exists', return_value=False)
    @mock.patch('builtins.open', side_effect=AssertionError('should not read'))
    def test_memoize_not_exists(self, mock_open, exists):
        # Reset the flag to appear that we haven't looked for it yet.
        netutils._IS_IPV6_ENABLED = None
        enabled = netutils.is_ipv6_enabled()
        self.assertFalse(enabled)
        enabled = netutils.is_ipv6_enabled()
        self.assertFalse(enabled)



#    Copyright (C) 2012-2013 Yahoo! Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Reflection module.

.. versionadded:: 1.1
"""

import inspect
import logging
import operator
import types
from typing import Any
from collections.abc import Generator, Sequence

try:
    _TYPE_TYPE = types.TypeType  # type: ignore
except AttributeError:
    _TYPE_TYPE = type

# See: https://docs.python.org/2/library/__builtin__.html#module-__builtin__
# and see https://docs.python.org/2/reference/executionmodel.html (and likely
# others)...
_BUILTIN_MODULES = ('builtins', '__builtin__', '__builtins__', 'exceptions')


LOG = logging.getLogger(__name__)


Parameter = inspect.Parameter
Signature = inspect.Signature
get_signature = inspect.signature


def get_members(
    obj: Any, exclude_hidden: bool = True
) -> Generator[tuple[str, Any], None, None]:
    """Yields the members of an object, filtering by hidden/not hidden.

    .. versionadded:: 2.3
    """
    for name, value in inspect.getmembers(obj):
        if name.startswith("_") and exclude_hidden:
            continue
        yield (name, value)


def get_member_names(obj: Any, exclude_hidden: bool = True) -> list[str]:
    """Get all the member names for a object."""
    return [
        name
        for (name, _obj) in get_members(obj, exclude_hidden=exclude_hidden)
    ]


def get_class_name(
    obj: Any, fully_qualified: bool = True, truncate_builtins: bool = True
) -> str:
    """Get class name for object.

    If object is a type, returns name of the type. If object is a bound
    method or a class method, returns its ``self`` object's class name.
    If object is an instance of class, returns instance's class name.
    Else, name of the type of the object is returned. If fully_qualified
    is True, returns fully qualified name of the type. For builtin types,
    just name is returned. TypeError is raised if can't get class name from
    object.
    """
    if inspect.isfunction(obj):
        raise TypeError("Can't get class name.")

    if inspect.ismethod(obj):
        obj = get_method_self(obj)
    if not isinstance(obj, type):
        obj = type(obj)
    if truncate_builtins:
        try:
            built_in = obj.__module__ in _BUILTIN_MODULES
        except AttributeError:  # nosec
            pass
        else:
            if built_in:
                return str(obj.__name__)
    if fully_qualified and hasattr(obj, '__module__'):
        return f'{obj.__module__}.{obj.__name__}'
    else:
        return str(obj.__name__)


def get_all_class_names(
    obj: Any,
    up_to: type = object,
    fully_qualified: bool = True,
    truncate_builtins: bool = True,
) -> Generator[str, None, None]:
    """Get class names of object parent classes.

    Iterate over all class names object is instance or subclass of,
    in order of method resolution (mro). If up_to parameter is provided,
    only name of classes that are sublcasses to that class are returned.
    """
    if not isinstance(obj, type):
        obj = type(obj)
    for cls in obj.mro():
        if issubclass(cls, up_to):
            yield get_class_name(
                cls,
                fully_qualified=fully_qualified,
                truncate_builtins=truncate_builtins,
            )


def get_callable_name(function: Any) -> str:
    """Generate a name from callable.

    Tries to do the best to guess fully qualified callable name.
    """
    method_self = get_method_self(function)
    parts: Sequence[str]
    if method_self is not None:
        # This is a bound method.
        if isinstance(method_self, type):
            # This is a bound class method.
            im_class = method_self
        else:
            im_class = type(method_self)
        try:
            parts = (im_class.__module__, function.__qualname__)
        except AttributeError:
            parts = (im_class.__module__, im_class.__name__, function.__name__)
    elif inspect.ismethod(function) or inspect.isfunction(function):
        # This could be a function, a static method, a unbound method...
        try:
            parts = (function.__module__, function.__qualname__)
        except AttributeError:
            if hasattr(function, 'im_class'):
                # This is a unbound method, which exists only in python 2.x
                im_class = function.im_class
                parts = (
                    im_class.__module__,
                    im_class.__name__,
                    function.__name__,
                )
            else:
                parts = (function.__module__, function.__name__)
    else:
        im_class = type(function)
        if im_class is _TYPE_TYPE:
            im_class = function
        try:
            parts = (im_class.__module__, im_class.__qualname__)
        except AttributeError:
            parts = (im_class.__module__, im_class.__name__)
    return '.'.join(parts)


def get_method_self(method: Any) -> Any:
    """Gets the ``self`` object attached to this method (or none)."""
    if not inspect.ismethod(method):
        return None
    try:
        return operator.attrgetter("__self__")(method)
    except AttributeError:
        return None


def is_same_callback(callback1: Any, callback2: Any) -> bool:
    """Returns if the two callbacks are the same."""
    if callback1 is callback2:
        # This happens when plain methods are given (or static/non-bound
        # methods).
        return True
    if callback1 == callback2:
        # NOTE(gmann): python3.8 onward, comparison of bound methods is
        # changed. It no longer decide the bound method's equality based
        # on their bounded objects equality instead it checks the identity
        # of their '__self__'.
        return True
    return False


def is_bound_method(method: Any) -> bool:
    """Returns if the given method is bound to an object."""
    return get_method_self(method) is not None


def is_subclass(obj: Any, cls: type) -> bool:
    """Returns if the object is class and it is subclass of a given class."""
    return inspect.isclass(obj) and issubclass(obj, cls)


def get_callable_args(function: Any, required_only: bool = False) -> list[str]:
    """Get names of callable arguments.

    Special arguments (like ``*args`` and ``**kwargs``) are not included into
    output.

    If required_only is True, optional arguments (with default values)
    are not included into output.
    """
    sig = get_signature(function)
    function_args = list(sig.parameters.keys())
    for param_name, p in sig.parameters.items():
        if p.kind in (Parameter.VAR_POSITIONAL, Parameter.VAR_KEYWORD) or (
            required_only and p.default is not Parameter.empty
        ):
            function_args.remove(param_name)
    return function_args


def accepts_kwargs(function: Any) -> bool:
    """Returns ``True`` if function accepts kwargs otherwise ``False``."""
    sig = get_signature(function)
    return any(
        p.kind == Parameter.VAR_KEYWORD for p in sig.parameters.values()
    )



# Copyright 2015 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime

from oslotest import base as test_base

from oslo_utils import fixture
from oslo_utils.fixture import keystoneidsentinel as keystids
from oslo_utils.fixture import uuidsentinel as uuids
from oslo_utils import timeutils
from oslo_utils import uuidutils


class TimeFixtureTest(test_base.BaseTestCase):
    def test_set_time_override_using_default(self):
        # When the fixture is used with its default constructor, the
        # override_time is set to the current timestamp.
        # Also, when the fixture is cleaned up, the override_time is reset.

        self.assertIsNone(timeutils.utcnow.override_time)
        with fixture.TimeFixture():
            self.assertIsNotNone(timeutils.utcnow.override_time)
        self.assertIsNone(timeutils.utcnow.override_time)

    def test_set_time_override(self):
        # When the fixture is used to set a time, utcnow returns that time.

        new_time = datetime.datetime(2015, 1, 2, 3, 4, 6, 7)
        self.useFixture(fixture.TimeFixture(new_time))
        self.assertEqual(new_time, timeutils.utcnow())
        # Call again to make sure it keeps returning the same time.
        self.assertEqual(new_time, timeutils.utcnow())

    def test_advance_time_delta(self):
        # advance_time_delta() advances the overridden time by some timedelta.

        new_time = datetime.datetime(2015, 1, 2, 3, 4, 6, 7)
        time_fixture = self.useFixture(fixture.TimeFixture(new_time))
        time_fixture.advance_time_delta(datetime.timedelta(seconds=1))
        expected_time = datetime.datetime(2015, 1, 2, 3, 4, 7, 7)
        self.assertEqual(expected_time, timeutils.utcnow())

    def test_advance_time_seconds(self):
        # advance_time_seconds() advances the overridden time by some number of
        # seconds.

        new_time = datetime.datetime(2015, 1, 2, 3, 4, 6, 7)
        time_fixture = self.useFixture(fixture.TimeFixture(new_time))
        time_fixture.advance_time_seconds(2)
        expected_time = datetime.datetime(2015, 1, 2, 3, 4, 8, 7)
        self.assertEqual(expected_time, timeutils.utcnow())


class UUIDSentinelsTest(test_base.BaseTestCase):
    def test_different_sentinel(self):
        uuid1 = uuids.foobar
        uuid2 = uuids.barfoo
        self.assertNotEqual(uuid1, uuid2)
        keystid1 = keystids.foobar
        keystid2 = keystids.barfoo
        self.assertNotEqual(keystid1, keystid2)

    def test_returns_uuid(self):
        self.assertTrue(uuidutils.is_uuid_like(uuids.foo))
        self.assertTrue(uuidutils.is_uuid_like(keystids.foo))

    def test_returns_string(self):
        self.assertIsInstance(uuids.foo, str)
        self.assertIsInstance(keystids.foo, str)

    def test_with_underline_prefix(self):
        ex = self.assertRaises(AttributeError, getattr, uuids, '_foo')
        self.assertIn("Sentinels must not start with _", str(ex))
        ex = self.assertRaises(AttributeError, getattr, keystids, '_foo')
        self.assertIn("Sentinels must not start with _", str(ex))



# Copyright 2011 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
import sys

from oslotest import base as test_base

from oslo_utils import importutils


class ImportUtilsTest(test_base.BaseTestCase):
    # NOTE(jkoelker) There has GOT to be a way to test this. But mocking
    #                __import__ is the devil. Right now we just make
    #               sure we can import something from the stdlib
    def test_import_class(self):
        dt = importutils.import_class('datetime.datetime')
        self.assertEqual(sys.modules['datetime'].datetime, dt)

    def test_import_bad_class(self):
        self.assertRaises(
            ImportError, importutils.import_class, 'lol.u_mad.brah'
        )

    def test_import_module(self):
        dt = importutils.import_module('datetime')
        self.assertEqual(sys.modules['datetime'], dt)

    def test_import_object_optional_arg_not_present(self):
        obj = importutils.import_object('oslo_utils.tests.fake.FakeDriver')
        self.assertEqual(obj.__class__.__name__, 'FakeDriver')

    def test_import_object_optional_arg_present(self):
        obj = importutils.import_object(
            'oslo_utils.tests.fake.FakeDriver', first_arg=False
        )
        self.assertEqual(obj.__class__.__name__, 'FakeDriver')

    def test_import_object_required_arg_not_present(self):
        # arg 1 isn't optional here
        self.assertRaises(
            TypeError,
            importutils.import_object,
            'oslo_utils.tests.fake.FakeDriver2',
        )

    def test_import_object_required_arg_present(self):
        obj = importutils.import_object(
            'oslo_utils.tests.fake.FakeDriver2', first_arg=False
        )
        self.assertEqual(obj.__class__.__name__, 'FakeDriver2')

    # namespace tests
    def test_import_object_ns_optional_arg_not_present(self):
        obj = importutils.import_object_ns(
            'oslo_utils', 'tests.fake.FakeDriver'
        )
        self.assertEqual(obj.__class__.__name__, 'FakeDriver')

    def test_import_object_ns_optional_arg_present(self):
        obj = importutils.import_object_ns(
            'oslo_utils', 'tests.fake.FakeDriver', first_arg=False
        )
        self.assertEqual(obj.__class__.__name__, 'FakeDriver')

    def test_import_object_ns_required_arg_not_present(self):
        # arg 1 isn't optional here
        self.assertRaises(
            TypeError,
            importutils.import_object_ns,
            'oslo_utils',
            'tests.fake.FakeDriver2',
        )

    def test_import_object_ns_required_arg_present(self):
        obj = importutils.import_object_ns(
            'oslo_utils', 'tests.fake.FakeDriver2', first_arg=False
        )
        self.assertEqual(obj.__class__.__name__, 'FakeDriver2')

    # namespace tests
    def test_import_object_ns_full_optional_arg_not_present(self):
        obj = importutils.import_object_ns(
            'tests2', 'oslo_utils.tests.fake.FakeDriver'
        )
        self.assertEqual(obj.__class__.__name__, 'FakeDriver')

    def test_import_object_ns_full_optional_arg_present(self):
        obj = importutils.import_object_ns(
            'tests2', 'oslo_utils.tests.fake.FakeDriver', first_arg=False
        )
        self.assertEqual(obj.__class__.__name__, 'FakeDriver')

    def test_import_object_ns_full_required_arg_not_present(self):
        # arg 1 isn't optional here
        self.assertRaises(
            TypeError,
            importutils.import_object_ns,
            'tests2',
            'oslo_utils.tests.fake.FakeDriver2',
        )

    def test_import_object_ns_full_required_arg_present(self):
        obj = importutils.import_object_ns(
            'tests2', 'oslo_utils.tests.fake.FakeDriver2', first_arg=False
        )
        self.assertEqual(obj.__class__.__name__, 'FakeDriver2')

    def test_import_object_ns_raise_import_error_in_init(self):
        self.assertRaises(
            ImportError,
            importutils.import_object_ns,
            'tests2',
            'oslo_utils.tests.fake.FakeDriver3',
        )

    def test_import_object(self):
        dt = importutils.import_object('datetime.time')
        self.assertIsInstance(dt, sys.modules['datetime'].time)

    def test_import_object_with_args(self):
        dt = importutils.import_object('datetime.datetime', 2012, 4, 5)
        self.assertIsInstance(dt, sys.modules['datetime'].datetime)
        self.assertEqual(dt, datetime.datetime(2012, 4, 5))

    def test_import_versioned_module(self):
        v2 = importutils.import_versioned_module('oslo_utils.tests.fake', 2)
        self.assertEqual(sys.modules['oslo_utils.tests.fake.v2'], v2)

        dummpy = importutils.import_versioned_module(
            'oslo_utils.tests.fake', 2, 'dummpy'
        )
        self.assertEqual(
            sys.modules['oslo_utils.tests.fake.v2.dummpy'], dummpy
        )

    def test_import_versioned_module_wrong_version_parameter(self):
        self.assertRaises(
            ValueError,
            importutils.import_versioned_module,
            'oslo_utils.tests.fake',
            "2.0",
            'fake',
        )

    def test_import_versioned_module_error(self):
        self.assertRaises(
            ImportError,
            importutils.import_versioned_module,
            'oslo_utils.tests.fake',
            2,
            'fake',
        )

    def test_try_import(self):
        dt = importutils.try_import('datetime')
        self.assertEqual(sys.modules['datetime'], dt)

    def test_try_import_returns_default(self):
        foo = importutils.try_import('foo.bar')
        self.assertIsNone(foo)

    def test_import_any_none_found(self):
        self.assertRaises(
            ImportError, importutils.import_any, 'foo.bar', 'foo.foo.bar'
        )

    def test_import_any_found(self):
        dt = importutils.import_any('foo.bar', 'datetime')
        self.assertEqual(sys.modules['datetime'], dt)



# Copyright 2011 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Import related utilities and helper functions.
"""

import sys
import traceback
from typing import Any


def import_class(import_str: str) -> Any:
    """Returns a class from a string including module and class.

    .. versionadded:: 0.3
    """
    mod_str, _sep, class_str = import_str.rpartition('.')
    __import__(mod_str)
    try:
        return getattr(sys.modules[mod_str], class_str)
    except AttributeError:
        raise ImportError(
            f'Class {class_str} cannot be found ({traceback.format_exception(*sys.exc_info())})'
        )


def import_object(import_str: str, *args: Any, **kwargs: Any) -> Any:
    """Import a class and return an instance of it.

    .. versionadded:: 0.3
    """
    return import_class(import_str)(*args, **kwargs)


def import_object_ns(
    name_space: str, import_str: str, *args: Any, **kwargs: Any
) -> Any:
    """Tries to import object from default namespace.

    Imports a class and return an instance of it, first by trying
    to find the class in a default namespace, then failing back to
    a full path if not found in the default namespace.

    .. versionadded:: 0.3

    .. versionchanged:: 2.6
       Don't capture :exc:`ImportError` when instanciating the object, only
       when importing the object class.
    """
    import_value = f"{name_space}.{import_str}"
    try:
        cls = import_class(import_value)
    except ImportError:
        cls = import_class(import_str)
    return cls(*args, **kwargs)


def import_module(import_str: str) -> Any:
    """Import a module.

    .. versionadded:: 0.3
    """
    __import__(import_str)
    return sys.modules[import_str]


def import_versioned_module(
    module: str, version: str | int, submodule: str | None = None
) -> Any:
    """Import a versioned module in format {module}.v{version][.{submodule}].

    :param module: the module name.
    :param version: the version number.
    :param submodule: the submodule name.
    :raises ValueError: For any invalid input.

    .. versionadded:: 0.3

    .. versionchanged:: 3.17
       Added *module* parameter.
    """

    # NOTE(gcb) Disallow parameter version include character '.'
    if '.' in f'{version}':
        raise ValueError("Parameter version shouldn't include character '.'.")
    module_str = f'{module}.v{version}'
    if submodule:
        module_str = '.'.join((module_str, submodule))
    return import_module(module_str)


def try_import(import_str: str, default: Any = None) -> Any:
    """Try to import a module and if it fails return default."""
    try:
        return import_module(import_str)
    except ImportError:
        return default


def import_any(module: str, *modules: str) -> Any:
    """Try to import a module from a list of modules.

    :param modules: A list of modules to try and import
    :returns: The first module found that can be imported
    :raises ImportError: If no modules can be imported from list

    .. versionadded:: 3.8
    """
    for module_name in (module,) + modules:
        imported_module = try_import(module_name)
        if imported_module:
            return imported_module

    raise ImportError(
        f'Unable to import any modules from the list {str(modules)}'
    )
